from __future__ import annotations
import asyncio
from collections.abc import Callable, Coroutine
from dataclasses import dataclass, field
import inspect
import logging
import sys
from typing import Any, Optional, TypeVar, TYPE_CHECKING, Union
from types import MethodType
import weakref

from .room.model import Room, RoomMessage
from .user.model import BlockingFlag, ChatMessage, TrackingState, User
from .protocol.primitives import (
    DirectoryData,
    MessageDataclass,
    Recommendation,
)
from .protocol.messages import (
    AddUser,
    AdminMessage,
    CheckPrivileges,
    GetGlobalRecommendations,
    GetItemRecommendations,
    GetItemSimilarUsers,
    GetRecommendations,
    GetSimilarUsers,
    GetUserInterests,
    GetUserStats,
    GetUserStatus,
    JoinRoom,
    Kicked,
    LeaveRoom,
    Login,
    PeerDirectoryContentsReply,
    PeerSharesReply,
    PeerUserInfoReply,
    PrivateChatMessage,
    PrivateRoomGrantMembership,
    PrivateRoomGrantOperator,
    PrivateRoomMembers,
    PrivateRoomMembershipGranted,
    PrivateRoomMembershipRevoked,
    PrivateRoomOperatorGranted,
    PrivateRoomOperatorRevoked,
    PrivateRoomOperators,
    PrivateRoomRevokeMembership,
    PrivateRoomRevokeOperator,
    PrivilegedUsers,
    PublicChatMessage,
    RoomChatMessage,
    RoomList,
    RoomTickerAdded,
    RoomTickerRemoved,
    RoomTickers,
    UserJoinedRoom,
    UserLeftRoom,
)
from .search.model import SearchRequest, SearchResult
from .shares.model import SharedDirectory
from .session import Session

if TYPE_CHECKING:
    from .network.connection import (
        Connection,
        ConnectionState,
        CloseReason,
        PeerConnection,
    )
    from .transfer.model import Transfer, TransferProgressSnapshot


logger = logging.getLogger(__name__)

# There used to be a difference in older versions between using the inspect and
# asyncio functions. The inspect function was fixed in 3.13. The asyncio version
# will be removed in 3.16
if sys.version_info >= (3, 14):
    iscoroutinefunction = inspect.iscoroutinefunction

else:
    iscoroutinefunction = asyncio.iscoroutinefunction


E = TypeVar('E', bound='Event')


EventListener = Union[
    Callable[[E], None],
    Callable[[E], Coroutine[Any, Any, Any]]
]


# Internal functions

def on_message(message_class: type[MessageDataclass]):
    """Decorator for methods listening to specific :class:`.MessageData` events
    """
    def register(event_func):
        event_func._registered_message = message_class
        return event_func

    return register


def build_message_map(obj: object) -> dict[type[MessageDataclass], Callable]:
    methods = inspect.getmembers(obj, predicate=inspect.ismethod)
    mapping: dict[type[MessageDataclass], Callable] = {}
    for _, method in methods:
        registered_message = getattr(method, '_registered_message', None)
        if registered_message:
            mapping[registered_message] = method
    return mapping


# External functions

class EventBus:

    def __init__(self):
        self._events: dict[
            type[Event],
            list[tuple[int, weakref.ReferenceType[EventListener]]]] = {}

    def register(self, event_class: type[E], listener: EventListener, priority: int = 100):
        """Registers an event listener to listen on an event class. The order in
        which the listeners are called can be managed using the ``priority``
        parameter
        """
        ref_factory: type[weakref.ReferenceType] = weakref.ref
        if isinstance(listener, MethodType):
            ref_factory = weakref.WeakMethod

        entry = (
            priority,
            ref_factory(listener, lambda ref: self._remove_callback(event_class, ref))
        )
        try:
            self._events[event_class].append(entry)
        except KeyError:
            self._events[event_class] = [entry, ]

        self._events[event_class].sort(key=lambda e: e[0])

    def unregister(self, event_class: type[E], listener: EventListener):
        """Unregisters the event listener from event bus"""
        if event_class not in self._events:
            return

        self._events[event_class] = [
            listener_def
            for listener_def in self._events[event_class]
            if listener != listener_def[1]()
        ]

    async def emit(self, event: Event):
        """Emit an event on the event bus. Listeners can be co-routines or
        regular functions

        :param event: the event instance to emit
        """
        for listener in self._get_listeners_for_event(event):
            try:
                if iscoroutinefunction(listener):
                    await listener(event)
                else:
                    listener(event)

            except Exception:
                logger.exception(
                    "exception notifying listener %r of event %r",
                    listener, event
                )

    def emit_sync(self, event: Event):
        """Emit an event on the event bus. Only listener functions are notified

        :param event: the event instance to emit
        """
        for listener in self._get_listeners_for_event(event):
            try:
                if iscoroutinefunction(listener):
                    logger.warning(
                        "attempted to emit synchronous event to an async listener : %r",
                        listener
                    )
                else:
                    listener(event)

            except Exception:
                logger.exception(
                    "exception notifying listener %r of event %r",
                    listener, event
                )

    def _get_listeners_for_event(self, event: Event) -> list[EventListener]:
        try:
            listeners_refs = self._events[event.__class__]
        except KeyError:
            return []

        listeners = []
        for _, listener_ref in listeners_refs:
            listener = listener_ref()
            # Should never be None, but checked for type compatibility
            if listener:  # pragma: no cover
                listeners.append(listener)

        return listeners

    def _remove_callback(
            self, event_class: type[E],
            listener_ref: weakref.ReferenceType[EventListener]):

        self._events[event_class] = [
            listener_def
            for listener_def in self._events[event_class]
            if listener_ref != listener_def[1]
        ]


# Public events
class Event:
    """Base class for events"""


@dataclass(frozen=True, slots=True)
class SessionInitializedEvent(Event):
    """Emitted after successful login"""
    session: Session
    raw_message: Login.Response


@dataclass(frozen=True, slots=True)
class SessionDestroyedEvent(Event):
    """Emitted after the login session has been destroyed (this is always after
    disconnect)
    """
    session: Session


@dataclass(frozen=True, slots=True)
class AdminMessageEvent(Event):
    """Emitted when a global admin message has been received"""
    message: str
    raw_message: AdminMessage.Response


@dataclass(frozen=True, slots=True)
class KickedEvent(Event):
    """Emitted when we are kicked from the server"""
    raw_message: Kicked.Response


@dataclass(frozen=True, slots=True)
class UserTrackingStateChangedEvent(Event):
    """Emitted when the tracking state of a user has changed.

    Possibly states for the ``state`` property will be defined in by the
    :class:`.TrackingState` flags
    """

    user: User
    state: TrackingState
    raw_message: Optional[AddUser.Response] = None


@dataclass(frozen=True, slots=True)
class UserTrackingEvent(Event):
    """Emitted when a user is now successfully by the library tracked. This will
    be triggered when:

    The ``user`` object will contain the tracked user. ``data`` will contain the
    data returned by the server

    .. deprecated:: 1.6
        Use :class:`.UserTrackingStateChangedEvent` instead
    """
    user: User
    raw_message: AddUser.Response


@dataclass(frozen=True, slots=True)
class UserTrackingFailedEvent(Event):
    """
    .. deprecated:: 1.6
        Use :class:`.UserTrackingStateChangedEvent` instead
    """
    username: str
    raw_message: AddUser.Response


@dataclass(frozen=True, slots=True)
class UserUntrackingEvent(Event):
    """Emitted when a user is no longer tracked by the library

    .. deprecated:: 1.6
        Use :class:`.UserTrackingStateChangedEvent` instead
    """
    user: User


@dataclass(frozen=True, slots=True)
class UserStatusUpdateEvent(Event):
    """Emitted when the server sends an update for the users status / privileges
    or a request for it
    """
    before: User
    current: User
    raw_message: GetUserStatus.Response


@dataclass(frozen=True, slots=True)
class UserStatsUpdateEvent(Event):
    """Emitted when the server send an update for the user's sharing statistics
    """
    before: User
    current: User
    raw_message: GetUserStats.Response


@dataclass(frozen=True, slots=True)
class UserInfoUpdateEvent(Event):
    """Emitted when user info has been received"""
    before: User
    current: User
    raw_message: PeerUserInfoReply.Request


@dataclass(frozen=True, slots=True)
class RoomListEvent(Event):
    rooms: list[Room]
    raw_message: RoomList.Response


@dataclass(frozen=True, slots=True)
class RoomMessageEvent(Event):
    message: RoomMessage
    raw_message: RoomChatMessage.Response


@dataclass(frozen=True, slots=True)
class RoomTickersEvent(Event):
    """Emitted when a list of tickers has been received for a room"""
    room: Room
    tickers: dict[str, str]
    raw_message: RoomTickers.Response


@dataclass(frozen=True, slots=True)
class RoomTickerAddedEvent(Event):
    """Emitted when a ticker has been added to the room by a user"""
    room: Room
    user: User
    ticker: str
    raw_message: RoomTickerAdded.Response


@dataclass(frozen=True, slots=True)
class RoomTickerRemovedEvent(Event):
    """Emitted when a ticker has been removed from the room by a user"""
    room: Room
    user: User
    raw_message: RoomTickerRemoved.Response


@dataclass(frozen=True, slots=True)
class RoomJoinedEvent(Event):
    """Emitted after a user joined a chat room

    The value of ``user`` will be ``None`` in case it is us who has joined the
    room
    """
    room: Room
    raw_message: Union[JoinRoom.Response, UserJoinedRoom.Response]
    user: Optional[User] = None


@dataclass(frozen=True, slots=True)
class RoomLeftEvent(Event):
    """Emitted after a user left a chat room

    The value of ``user`` will be ``None`` in case it is us who has left the
    room
    """
    room: Room
    raw_message: Union[LeaveRoom.Response, UserLeftRoom.Response]
    user: Optional[User] = None


@dataclass(frozen=True, slots=True)
class RoomMembershipGrantedEvent(Event):
    """Emitted when a member has been added to the private room

    The value of ``member`` will be ``None`` in case it is us who has been added
    to the room
    """
    room: Room
    raw_message: Union[PrivateRoomGrantMembership.Response, PrivateRoomMembershipGranted.Response]
    member: Optional[User] = None


@dataclass(frozen=True, slots=True)
class RoomMembershipRevokedEvent(Event):
    """Emitted when a member has been removed to the private room

    The value of ``member`` will be ``None`` in case it is us who has been
    removed from the room
    """
    room: Room
    raw_message: Union[PrivateRoomRevokeMembership.Response, PrivateRoomMembershipRevoked.Response]
    member: Optional[User] = None


@dataclass(frozen=True, slots=True)
class RoomOperatorGrantedEvent(Event):
    """Emitted when a member has been granted operator privileges on a private
    room

    The value of ``member`` will be ``None`` in case it is us who has been
    granted operator
    """
    room: Room
    raw_message: Union[PrivateRoomGrantOperator.Response, PrivateRoomOperatorGranted.Response]
    member: Optional[User] = None


@dataclass(frozen=True, slots=True)
class RoomOperatorRevokedEvent(Event):
    """Emitted when a member had operator privileges revoked on a private room

    The value of ``user`` will be ``None`` in case it is us who has been revoked
    operator
    """
    room: Room
    raw_message: Union[PrivateRoomRevokeOperator.Response, PrivateRoomOperatorRevoked.Response]
    member: Optional[User] = None


@dataclass(frozen=True, slots=True)
class RoomOperatorsEvent(Event):
    """Emitted when the server sends us a list of operators in a room"""
    room: Room
    operators: list[User]
    raw_message: PrivateRoomOperators.Response


@dataclass(frozen=True, slots=True)
class RoomMembersEvent(Event):
    """Emitted when the server sends us a list of members of a room. The list
    of members always excludes the owner of the room
    """
    room: Room
    members: list[User]
    raw_message: PrivateRoomMembers.Response


@dataclass(frozen=True, slots=True)
class PrivateMessageEvent(Event):
    """Emitted when a private message has been received"""
    message: ChatMessage
    raw_message: PrivateChatMessage.Response


@dataclass(frozen=True, slots=True)
class PublicMessageEvent(Event):
    """Emitted when a message has been received from a public room and public
    chat has been enabled
    """
    timestamp: int
    user: User
    room: Room
    message: str
    raw_message: PublicChatMessage.Response


@dataclass(frozen=True, slots=True)
class SearchRequestSentEvent(Event):
    """Emitted when a search request has been sent out"""
    query: SearchRequest


@dataclass(frozen=True, slots=True)
class SearchRequestRemovedEvent(Event):
    """Emitted when a search request has been removed"""
    query: SearchRequest


@dataclass(frozen=True, slots=True)
class SearchResultEvent(Event):
    """Emitted when a search result has been received"""
    query: SearchRequest
    result: SearchResult


@dataclass(frozen=True, slots=True)
class SearchRequestReceivedEvent(Event):
    """Emitted when a search request by another user has been received"""
    username: str
    query: str
    result_count: int


@dataclass(frozen=True, slots=True)
class SimilarUsersEvent(Event):
    users: list[tuple[User, int]]
    raw_message: GetSimilarUsers.Response


@dataclass(frozen=True, slots=True)
class ItemSimilarUsersEvent(Event):
    item: str
    users: list[User]
    raw_message: GetItemSimilarUsers.Response


@dataclass(frozen=True, slots=True)
class RecommendationsEvent(Event):
    recommendations: list[Recommendation]
    unrecommendations: list[Recommendation]
    raw_message: GetRecommendations.Response


@dataclass(frozen=True, slots=True)
class GlobalRecommendationsEvent(Event):
    recommendations: list[Recommendation]
    unrecommendations: list[Recommendation]
    raw_message: GetGlobalRecommendations.Response


@dataclass(frozen=True, slots=True)
class ItemRecommendationsEvent(Event):
    item: str
    recommendations: list[Recommendation]
    raw_message: GetItemRecommendations.Response


@dataclass(frozen=True, slots=True)
class UserInterestsEvent(Event):
    user: User
    interests: list[str]
    hated_interests: list[str]
    raw_message: GetUserInterests.Response


@dataclass(frozen=True, slots=True)
class PrivilegedUsersEvent(Event):
    """Emitted when the list of privileged users has been received"""
    users: list[User]
    raw_message: PrivilegedUsers.Response


@dataclass(frozen=True, slots=True)
class PrivilegedUserAddedEvent(Event):
    """Emitted when a new privileged user has been added"""
    user: User


@dataclass(frozen=True, slots=True)
class PrivilegesUpdateEvent(Event):
    """Emitted when we receive a message containing how much time is left on the
    current user's privileges. If ``time_left`` is 0 the current user has no
    privileges
    """
    time_left: int
    raw_message: CheckPrivileges.Response


# Peer

@dataclass(frozen=True, slots=True)
class UserSharesReplyEvent(Event):
    user: User
    directories: list[DirectoryData]
    locked_directories: list[DirectoryData]
    raw_message: PeerSharesReply.Request


@dataclass(frozen=True, slots=True)
class UserDirectoryEvent(Event):
    user: User
    directory: str
    directories: list[DirectoryData]
    raw_message: PeerDirectoryContentsReply.Request


# Transfer

@dataclass(frozen=True, slots=True)
class TransferAddedEvent(Event):
    """Emitted when a transfer has been attached to the client"""
    transfer: Transfer


@dataclass(frozen=True, slots=True)
class TransferRemovedEvent(Event):
    """Emitted when a transfer has been detached from the client"""
    transfer: Transfer


@dataclass(frozen=True, slots=True)
class TransferProgressEvent(Event):
    """Called periodically to report progress. The interval is determined by the
    ``transfers.progress_interval`` settings parameter

    This event only includes transfers where the transfer state has been changed
    since the previous event. If there are no updates the event will no be
    called
    """
    updates: list[tuple[Transfer, TransferProgressSnapshot, TransferProgressSnapshot]]
    """List of progress updates: transfer instance, previous snapshot, current
    snapshot
    """


# Internal Events

class InternalEvent(Event):
    """Base class for internal events"""


@dataclass(frozen=True, slots=True)
class ConnectionStateChangedEvent(InternalEvent):
    """Indicates a state change in any type of connection"""
    connection: Connection
    state: ConnectionState
    close_reason: Optional[CloseReason] = None


@dataclass(frozen=True, slots=True)
class ServerReconnectedEvent(InternalEvent):
    """Indicates server reconnected, used internally to automatically log back
    in
    """


@dataclass(frozen=True, slots=True)
class MessageReceivedEvent(InternalEvent):
    """Emitted when a message was received on a connection"""
    message: MessageDataclass
    connection: Connection


@dataclass(frozen=True, slots=True)
class PeerInitializedEvent(InternalEvent):
    """Emitted when a new peer connection has been established and the
    initialization message has been received or sent
    """
    connection: PeerConnection
    requested: bool
    """Indictes whether the connection was initialized by another user or opened
    on the client's request
    """


@dataclass(frozen=True, slots=True)
class ScanCompleteEvent(InternalEvent):
    """Emitted when shares scan was completed"""
    folder_count: int
    file_count: int


@dataclass(frozen=True, slots=True)
class FriendListChangedEvent(InternalEvent):
    """Emitted when a change was detected in the friends list"""
    added: set[str] = field(default_factory=set)
    removed: set[str] = field(default_factory=set)


@dataclass(frozen=True, slots=True)
class BlockListChangedEvent(InternalEvent):
    """Emitted when a change was detected in the blocked list"""
    changes: dict[str, tuple[BlockingFlag, BlockingFlag]]
    """List of changes. The key contains the username, the value is a tuple
    containing the old and new flags.
    """


@dataclass(frozen=True, slots=True)
class SharedDirectoryChangeEvent(InternalEvent):
    shared_directory: SharedDirectory



from __future__ import annotations
import abc
from aiofiles.threadpool.binary import AsyncBufferedIOBase, AsyncBufferedReader
import asyncio
from async_timeout import Timeout, timeout as atimeout
from collections.abc import Awaitable, Callable
from enum import auto, Enum
from typing import Optional, TYPE_CHECKING, Union
import logging
import socket
import struct

from ..constants import (
    DEFAULT_READ_TIMEOUT,
    DISCONNECT_TIMEOUT,
    PEER_CONNECT_TIMEOUT,
    PEER_READ_TIMEOUT,
    SERVER_CONNECT_TIMEOUT,
    SERVER_READ_TIMEOUT,
    TRANSFER_TIMEOUT,
)
from ..exceptions import (
    ConnectionFailedError,
    ConnectionReadError,
    ConnectionWriteError,
    MessageDeserializationError,
    MessageSerializationError,
)
from ..protocol import obfuscation
from ..protocol.primitives import uint32, uint64, MessageDataclass
from ..protocol.messages import (
    DistributedMessage,
    PeerInitializationMessage,
    PeerMessage,
    ServerMessage,
)
from ..log_utils import ConnectionLoggerAdapter
from ..utils import task_counter
from .rate_limiter import RateLimiter, UnlimitedRateLimiter

if TYPE_CHECKING:
    from .network import Network


logger = logging.getLogger(__name__)
adapter = ConnectionLoggerAdapter(logger, {})


HEADER_SIZE_OBFUSCATED: int = obfuscation.KEY_SIZE + struct.calcsize('I')
HEADER_SIZE_UNOBFUSCATED: int = struct.calcsize('I')


class PeerConnectionType:
    FILE = 'F'
    PEER = 'P'
    DISTRIBUTED = 'D'


class ConnectionState(Enum):
    UNINITIALIZED = auto()
    CONNECTING = auto()
    CONNECTED = auto()
    CLOSING = auto()
    CLOSED = auto()


class CloseReason(Enum):
    UNKNOWN = auto()
    CONNECT_FAILED = auto()
    REQUESTED = auto()
    READ_ERROR = auto()
    WRITE_ERROR = auto()
    TIMEOUT = auto()
    EOF = auto()


class PeerConnectionState(Enum):
    AWAITING_INIT = auto()
    """No init message has been received yet (PeerInit, PeerPierceFirewall)"""
    ESTABLISHED = auto()
    """Message connections: ready to receive / send data"""
    NEGOTIATING_TRANSFER = auto()
    """Transfer connections: negotiating the transfer ticket / offset"""
    TRANSFERRING = auto()
    """Transfer connections: ready to receive / send data"""


class Connection(abc.ABC):

    _CLOSING_STATES: tuple[ConnectionState, ConnectionState] = (
        ConnectionState.CLOSING, ConnectionState.CLOSED)

    def __init__(self, hostname: str, port: int, network: Network):
        self.hostname: str = hostname
        self.port: int = port
        self.network: Network = network
        self.state: ConnectionState = ConnectionState.UNINITIALIZED
        self._is_closing: bool = False

    @abc.abstractmethod
    async def disconnect(self, reason: CloseReason = CloseReason.UNKNOWN):
        ...

    async def set_state(self, state: ConnectionState, close_reason: CloseReason = CloseReason.UNKNOWN):
        self.state = state
        self._is_closing = state in self._CLOSING_STATES
        await self.network.on_state_changed(state, self, close_reason=close_reason)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(hostname={self.hostname!r}, port={self.port}, "
            f"state={self.state})")


class ListeningConnection(Connection):
    """A listening connection, objects of this class are responsible for
    accepting incoming connections from peers
    """

    def __init__(self, hostname: str, port: int, network: Network, obfuscated: bool = False):
        super().__init__(hostname, port, network)
        self.obfuscated: bool = obfuscated

        self._server: Optional[asyncio.AbstractServer] = None

    async def disconnect(self, reason: CloseReason = CloseReason.UNKNOWN):
        adapter.debug("disconnecting : %s", reason.name, extra=self.__dict__)
        await self.set_state(ConnectionState.CLOSING, close_reason=reason)
        try:
            if self._server is not None:
                if self._server.is_serving():
                    self._server.close()

                async with atimeout(DISCONNECT_TIMEOUT):
                    await self._server.wait_closed()

        except Exception as exc:
            adapter.warning(
                "exception while disconnecting", extra=self.__dict__, exc_info=exc)

        finally:
            self._server = None
            adapter.info("disconnected : %s", reason.name, extra=self.__dict__)
            await self.set_state(ConnectionState.CLOSED, close_reason=reason)

    async def connect(self):
        """Open a listening connection on the current :attr:`hostname` and
        :attr:`port`

        :raise ConnectionFailedError: raised when binding failed
        """
        adapter.info("open listening connection", extra=self.__dict__)
        await self.set_state(ConnectionState.CONNECTING)

        try:
            self._server = await asyncio.start_server(
                self.accept,
                self.hostname,
                self.port,
                family=socket.AF_INET,
                start_serving=True
            )

        except OSError as exc:
            adapter.exception("failed to bind listening port", extra=self.__dict__)
            await self.disconnect(CloseReason.CONNECT_FAILED)
            raise ConnectionFailedError(f"{self.hostname}:{self.port} : failed to connect") from exc

        await self.set_state(ConnectionState.CONNECTED)

    async def accept(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter):
        hostname, port = writer.get_extra_info('peername')

        adapter.debug(
            "accepted connection : %s:%d", hostname, port, extra=self.__dict__)

        connection = PeerConnection(
            hostname, port, self.network,
            obfuscated=self.obfuscated,
            connection_type=PeerConnectionType.PEER,
            incoming=True
        )
        connection._reader, connection._writer = reader, writer
        await self.network.on_peer_accepted(connection)
        await connection.set_state(ConnectionState.CONNECTED)


class DataConnection(Connection, abc.ABC):
    """Connection for message and data transfer"""

    def __init__(
            self, hostname: str, port: int, network: Network,
            obfuscated: bool = False, read_timeout: float = DEFAULT_READ_TIMEOUT):

        super().__init__(hostname, port, network)
        self.obfuscated = obfuscated

        self._reader: Optional[asyncio.StreamReader] = None
        self._writer: Optional[asyncio.StreamWriter] = None
        self._reader_task: Optional[asyncio.Task] = None

        self._queued_messages: list[asyncio.Task] = []
        self._read_timeout_object: Optional[Timeout] = None

        self.read_timeout: float = read_timeout

    def get_connecting_ip(self) -> str:
        """Gets the IP address being used to connect to the server/peer.

        The connection needs to be established for this method to work
        """
        if not self._writer:  # pragma: no cover
            raise RuntimeError("connection not initialized")

        return self._writer.get_extra_info('sockname')[0]

    async def connect(self, timeout: float = 30):
        """Opens a TCP connection the hostname:ip provided in this object. The
        state of the connection will be changed to ``CONNECTING`` before the
        attempt is made

        Upon success the :attr:`state` will be set to ``CONNECTED`` and the
        message reader loop will be started. Upon failure the state will be set
        to ``CLOSED`` with ``CONNECT_FAILED`` as its reason

        :param timeout: timeout in seconds before giving up (default: 30)
        :raise ConnectionFailedError: raised when connection failed or timed out
        """
        adapter.info("connecting", extra=self.__dict__)
        await self.set_state(ConnectionState.CONNECTING)

        try:
            async with atimeout(timeout):
                self._reader, self._writer = await asyncio.open_connection(
                    self.hostname, self.port)

        except (Exception, asyncio.TimeoutError) as exc:
            await self.disconnect(CloseReason.CONNECT_FAILED)
            raise ConnectionFailedError(f"{self.hostname}:{self.port} : failed to connect") from exc

        else:
            adapter.debug("connected", extra=self.__dict__)
            await self.set_state(ConnectionState.CONNECTED)

    async def disconnect(self, reason: CloseReason = CloseReason.UNKNOWN):
        """Disconnects the TCP connection. The method will not raise an
        exception in case the connection wasn't yet connected

        :param reason: optional reason for the disconnect. This parameter is
            purely informational
        """
        # Prevents this method being called twice during the process of
        # disconnecting. This is necessary because during disconnecting the
        # reader task throw an EOF and attempt to call this method again.
        if self.state in (ConnectionState.CLOSED, ConnectionState.CLOSING):
            return

        await self.set_state(ConnectionState.CLOSING, close_reason=reason)
        adapter.debug("disconnecting : %s", reason.name, extra=self.__dict__)
        self._cancel_queued_messages()
        try:
            if self._writer is not None:
                if not self._writer.is_closing():
                    self._writer.close()

                async with atimeout(DISCONNECT_TIMEOUT):
                    await self._writer.wait_closed()

        except Exception as exc:
            adapter.warning(
                "exception while disconnecting : %r", exc, extra=self.__dict__)

        finally:
            await self.set_state(ConnectionState.CLOSED, close_reason=reason)
            # Because disconnect can be called when read failed setting the
            # reader task to none should be done last
            self._reader_task = None
            self._reader = None
            self._writer = None

    def start_reader_task(self):
        """Starts the message reader task"""
        self._reader_task = asyncio.create_task(self._message_reader_loop())

    def stop_reader_task(self):
        """Stops the message reader task if it exists"""
        if self._reader_task is not None:
            self._reader_task.cancel()
            self._reader_task = None

    # Read/write methods

    async def _message_reader_loop(self):
        """Message reader loop. This will loop until the connection is closed or
        the network is closed
        """
        while not self._is_closing:
            try:
                message = await self.receive_message_object()

            except ConnectionReadError:
                adapter.warning("read error", extra=self.__dict__)

            except MessageDeserializationError as exc:
                adapter.warning(
                    "failed to deserialize message : %s", exc.proto_message, extra=self.__dict__)

            else:

                if not message:
                    return

                # Do not handle messages when closing/closed
                if self._is_closing:
                    adapter.warning(
                        "connection is closing, skipping handling message : %s",
                        message,
                        extra=self.__dict__
                    )
                else:
                    await self._perform_message_callback(message)

    async def _read(
            self, reader_coro: Awaitable[Optional[bytes]],
            timeout: Optional[float] = None) -> Optional[bytes]:
        """Read data from the connection using the passed ``reader_coro``. When
        an error occurs during reading the connection will be CLOSED

        :param reader_coro: coroutine that reads the data
        :return: the return value of the ``reader_coro``, ``None`` if EOF
        :raise ConnectionReadError: upon any kind of read error/timeout
        """
        try:
            if timeout:
                # TODO: change the type of `_read_timeout_object` to `asyncio.timeouts.Timeout`
                # when 3.11 is the oldest supported Python version
                async with atimeout(timeout) as self._read_timeout_object:  # type: ignore[assignment]
                    data = await reader_coro
            else:
                data = await reader_coro

        except asyncio.IncompleteReadError as exc:
            # Can only be raised by `readexactly`
            if exc.partial:
                await self.disconnect(CloseReason.READ_ERROR)
                raise ConnectionReadError(
                    f"{self.hostname}:{self.port} : incomplete read on connection : {exc.partial!r}") from exc
            else:
                await self.disconnect(CloseReason.EOF)
                return None

        except asyncio.TimeoutError as exc:
            await self.disconnect(CloseReason.TIMEOUT)
            raise ConnectionReadError(f"{self.hostname}:{self.port} : read timeout") from exc

        except Exception as exc:
            await self.disconnect(CloseReason.READ_ERROR)
            raise ConnectionReadError(f"{self.hostname}:{self.port} : exception during reading") from exc

        else:
            if not data:
                await self.disconnect(CloseReason.EOF)
                return None
            else:
                return data

    async def _read_message(self) -> bytes:
        if not self._reader:
            raise ConnectionReadError(
                f"{self.hostname}:{self.port} : cannot read message, connection is not open")

        header_size = HEADER_SIZE_OBFUSCATED if self.obfuscated else HEADER_SIZE_UNOBFUSCATED

        header = await self._reader.readexactly(header_size)

        message_len_buf = obfuscation.decode(header) if self.obfuscated else header
        _, message_len = uint32.deserialize(0, message_len_buf)

        message = await self._reader.readexactly(message_len)

        return header + message

    async def receive_message(self) -> Optional[bytes]:
        """Receives a single raw message

        :return: The message as a ``bytes`` object. ``None`` if the connection
            returned EOF
        """
        return await self._read(self._read_message(), timeout=self.read_timeout)

    async def receive_message_object(self) -> Optional[MessageDataclass]:
        """Receives a single message and parses it to a :class:`.MessageDataclass`

        :return: The parsed message, ``None`` if the connection returned EOF
        """
        message_data = await self.receive_message()

        if message_data:
            message = self.decode_message_data(message_data)
            adapter.debug(
                "received message : %s",
                message,
                extra={
                    **self.__dict__,
                    **{'message_type': message.__class__}
                }
            )
            return message

        return None

    async def receive_until_eof(self, raise_exception: bool = True) -> Optional[bytes]:
        """Receives data until the other end closes the connection. If
        ``raise_exception`` parameter is set to ``True`` other read errors are
        treated as EOF as well
        """
        if not self._reader:
            raise ConnectionReadError(
                f"{self.hostname}:{self.port} : cannot read until EOF, connection is not open")

        # Note: _read raises ConnectionReadError also on IncompleteReadError,
        # however when using StreamReader.read this error can never be raised
        # and thus this method shouldn't worry about potentially not returning
        # partial data
        try:
            return await self._read(self._reader.read(-1))
        except ConnectionReadError:
            if raise_exception:
                raise

            return b''

    def queue_message(self, message: Union[bytes, MessageDataclass]) -> asyncio.Task:
        task = asyncio.create_task(
            self.send_message(message),
            name=f'queue-message-task-{task_counter()}'
        )
        self._queued_messages.append(task)
        task.add_done_callback(self._queued_messages.remove)
        return task

    def queue_messages(self, *messages: Union[bytes, MessageDataclass]) -> list[asyncio.Task]:
        return [
            self.queue_message(message)
            for message in messages
        ]

    async def _send(self, data: bytes, timeout: Optional[float] = None):
        if not self._writer:
            raise ConnectionWriteError(
                f"{self.hostname}:{self.port} : cannot send data, connection is not open")

        try:
            self._writer.write(data)
            if timeout:
                async with atimeout(timeout):
                    await self._writer.drain()
            else:
                await self._writer.drain()

        except asyncio.TimeoutError as exc:
            await self.disconnect(CloseReason.TIMEOUT)
            raise ConnectionWriteError(f"{self.hostname}:{self.port} : write timeout") from exc

        except Exception as exc:
            await self.disconnect(CloseReason.WRITE_ERROR)
            raise ConnectionWriteError(f"{self.hostname}:{self.port} : exception during writing") from exc

    async def send_message(self, message: Union[bytes, MessageDataclass]):
        """Sends a message or a set of bytes over the connection. In case an
        object of :class:`.MessageDataClass` is provided the object will first
        be serialized. If the :attr:`obfuscated` flag is set for the connection
        the message or bytes will first be obfuscated

        :param message: message to be sent over the connection
        :raise ConnectionWriteError: error or timeout occured during writing
        """
        if self._is_closing:
            adapter.warning(
                "not sending message, connection is closing / closed : %s",
                message,
                extra=self.__dict__
            )
            return

        adapter.debug(
            "send message : %s",
            message,
            extra={
                **self.__dict__,
                **{'message_type': message.__class__}
            }
        )
        # Serialize the message
        try:
            data = self.encode_message_data(message)
        except MessageSerializationError:
            adapter.exception("failed to serialize message : %s", message)
            return

        await self._send(data, timeout=10)
        self._increase_read_timeout()

    def encode_message_data(self, message: Union[bytes, MessageDataclass]) -> bytes:
        """Serializes the :class:`.MessageDataclass` or ``bytes`` and obfuscates
        the contents. See :meth:`serialize_message`

        :return: ``bytes`` object
        :raise MessageSerializationError: raised when serialization failed
        """
        try:
            data = self.serialize_message(message)
        except Exception as exc:
            raise MessageSerializationError("failed to serialize message") from exc

        if self.obfuscated:
            data = obfuscation.encode(data)

        return data

    def decode_message_data(self, data: bytes) -> MessageDataclass:
        """De-obfuscates and deserializes message data into a
        :class:`.MessageDataclass` object. See :meth:`deserialize_message`

        :param data: message bytes to decode
        :return: object of :class:`.MessageDataclass`
        :raise MessageDeserializationError: raised when deserialization failed
        """
        if self.obfuscated:
            data = obfuscation.decode(data)

        try:
            message = self.deserialize_message(data)
        except Exception as exc:
            raise MessageDeserializationError(data, "failed to deserialize message") from exc

        return message

    @abc.abstractmethod
    def deserialize_message(self, message_data: bytes) -> MessageDataclass:
        """Should be called after a full message has been received. This method
        should parse the message
        """

    def serialize_message(self, message: Union[bytes, MessageDataclass]) -> bytes:
        if isinstance(message, MessageDataclass):
            return message.serialize()
        else:
            return message

    def _increase_read_timeout(self):
        if self.read_timeout and self._read_timeout_object:
            try:
                self._read_timeout_object.shift(self.read_timeout)
            except RuntimeError:
                pass

    async def _perform_message_callback(self, message: MessageDataclass):
        try:
            await self.network.on_message_received(message, self)
        except Exception:
            adapter.exception("error during callback : %s", message, extra=self.__dict__)

    def _cancel_queued_messages(self):
        for qmessage_task in self._queued_messages:
            qmessage_task.cancel()


class ServerConnection(DataConnection):

    def __init__(
            self, hostname: str, port: int, network: Network,
            obfuscated: bool = False, read_timeout: float = SERVER_READ_TIMEOUT):

        super().__init__(
            hostname, port, network,
            obfuscated=obfuscated, read_timeout=read_timeout)

    async def connect(self, timeout: float = SERVER_CONNECT_TIMEOUT):
        await super().connect(timeout=timeout)

    def deserialize_message(self, message_data: bytes) -> MessageDataclass:
        return ServerMessage.deserialize_response(message_data)


class PeerConnection(DataConnection):

    def __init__(
            self, hostname: str, port: int, network: Network, obfuscated: bool = False,
            username: Optional[str] = None, connection_type: str = PeerConnectionType.PEER,
            incoming: bool = False, read_timeout: float = PEER_READ_TIMEOUT,
            transfer_read_timeout: float = TRANSFER_TIMEOUT):

        super().__init__(
            hostname, port, network,
            obfuscated=obfuscated, read_timeout=read_timeout)

        self.transfer_read_timeout: float = transfer_read_timeout

        self.incoming: bool = incoming
        self.connection_state = PeerConnectionState.AWAITING_INIT
        self.connection_type: str = connection_type

        self.username: Optional[str] = username

        self.download_rate_limiter: RateLimiter = UnlimitedRateLimiter()
        self.upload_rate_limiter: RateLimiter = UnlimitedRateLimiter()

    async def connect(self, timeout: float = PEER_CONNECT_TIMEOUT):
        await super().connect(timeout=timeout)

    def set_connection_state(self, state: PeerConnectionState):
        """Sets the current connection state.

        If the current state is ``AWAITING_INIT`` and the connection type is a
        distributed or file connection the :attr:`obfuscated` flag for this
        connection will be set to ``False``

        If the state goes to the ``ESTABLISHED`` state the message reader task
        will be started. In all other cases it will be stopped (in case the
        task was running)

        :param state: The new state of the connection
        """
        # Set non-peer connections to non-obfuscated
        if state != PeerConnectionState.AWAITING_INIT:
            if self.connection_type != PeerConnectionType.PEER:
                self.obfuscated = False

        if state == PeerConnectionState.ESTABLISHED:
            self.start_reader_task()
        else:
            # This shouldn't occur, during all other states we call the
            # receive_* methods directly. But it can do no harm
            self.stop_reader_task()

        adapter.debug("setting state to %s", state, extra=self.__dict__)
        self.connection_state = state

    async def receive_transfer_ticket(self) -> int:
        """Receive the transfer ticket from the connection"""
        if not self._reader:
            raise ConnectionReadError(
                f"{self.hostname}:{self.port} : cannot read transfer ticket, "
                "connection is not open"
            )

        data = await self._read(self._reader.readexactly(struct.calcsize('I')))
        if data is None:  # pragma: no cover
            raise ConnectionReadError(
                f"{self.hostname}:{self.port} : couldn't receive transfer ticket, "
                "connection closed"
            )

        _, ticket = uint32.deserialize(0, data)
        return ticket

    async def receive_transfer_offset(self) -> int:
        """Receive the transfer offset from the connection"""
        if not self._reader:
            raise ConnectionReadError(
                f"{self.hostname}:{self.port} : cannot read transfer offset, "
                "connection is not open"
            )

        data = await self._read(self._reader.readexactly(struct.calcsize('Q')))
        if data is None:  # pragma: no cover
            raise ConnectionReadError(
                f"{self.hostname}:{self.port} : couldn't receive transfer offset, "
                "connection closed"
            )

        _, offset = uint64.deserialize(0, data)
        return offset

    async def receive_data(self, n_bytes: int) -> Optional[bytes]:
        """Receives data on the connection. The ``n_bytes`` indicates the max
        amount of bytes that should be received on the connection, less bytes
        can be returned.

        In case of error the socket will be disconnected. If no data is received
        it is assumed to be EOF and the connection will be disconnected.

        :param n_bytes: max amount of bytes to receive
        :raise ConnectionReadError: in case timeout occured on an error on the
            socket
        :return: ``bytes`` object containing the received data or ``None`` in
            case the connection reached EOF
        """
        if not self._reader:
            raise ConnectionReadError(
                f"{self.hostname}:{self.port} : cannot read data, connection is not open")

        return await self._read(
            self._reader.read(n_bytes), timeout=self.transfer_read_timeout)

    async def receive_file(
            self, file_handle: AsyncBufferedIOBase, filesize: int,
            callback: Optional[Callable[[bytes], None]] = None):
        """Receives a file on the current connection and writes it to the given
        ``file_handle``

        This method will attempt to keep reading data until EOF is received from
        the connection or the amount of received bytes has reached the
        ``filesize``

        :param file_handle: a file handle to write the received data to
        :param filesize: expected filesize
        :param callback: optional callback that gets called each time a chunk
            of data is received
        """
        bytes_received = 0
        while True:
            bytes_to_read = await self.download_rate_limiter.take_tokens()
            data = await self.receive_data(bytes_to_read)
            if data is None:
                return

            await file_handle.write(data)
            if callback is not None:
                callback(data)

            # Check if all data received and return
            bytes_received += len(data)
            if bytes_received >= filesize:
                return

    async def send_data(self, data: bytes):
        await self._send(data, timeout=TRANSFER_TIMEOUT)

    async def send_file(self, file_handle: AsyncBufferedReader, callback: Optional[Callable[[bytes], None]] = None):
        """Sends a file over the connection. This method makes use of the
        :attr:`upload_rate_limiter` to limit how many bytes are being sent at a
        time

        :param file_handle: binary opened file handle
        :param callback: progress callback that gets called each time a chunk
            of data was sent
        """
        while True:
            bytes_to_write = await self.upload_rate_limiter.take_tokens()
            data = await file_handle.read(bytes_to_write)
            if not data:
                return

            await self.send_data(data)
            if callback is not None:
                callback(data)

    def deserialize_message(self, message_data: bytes) -> MessageDataclass:
        if self.connection_state == PeerConnectionState.AWAITING_INIT:
            return PeerInitializationMessage.deserialize_request(message_data)

        else:
            if self.connection_type == PeerConnectionType.PEER:
                return PeerMessage.deserialize_request(message_data)
            else:
                return DistributedMessage.deserialize_request(message_data)

    def __repr__(self):
        return (
            f"{self.__class__.__name__}("
            f"hostname={self.hostname!r}, port={self.port}, state={self.state}, "
            f"incoming={self.incoming}, obfuscated={self.obfuscated}, username={self.username!r}, "
            f"connection_state={self.connection_state}, connection_type={self.connection_type!r})")



from __future__ import annotations
from dataclasses import dataclass, field
from enum import auto, Enum, Flag, IntFlag
import sys
from typing import Optional

from ..protocol.primitives import UserStats


if sys.version_info >= (3, 11):
    slots_params = {'slots': True, 'weakref_slot': True}

else:
    slots_params = {}


class UserStatus(Enum):
    """User status values, everything except the UNKNOWN status are used by the
    server
    """
    UNKNOWN = -1
    OFFLINE = 0
    AWAY = 1
    ONLINE = 2


class UploadPermissions(Enum):
    """Upload permissions returned / sent by PeerUserInfoReply message"""
    UNKNOWN = -1
    NOONE = 0
    """Don't allow uploads from anyone"""
    EVERYONE = 1
    """Allow uploads from anyone"""
    USER_LIST = 2
    """Only allow uploads from users in the users list"""
    PERMITTED_LIST = 3
    """Only allow uploads from users in a specific list"""


class TrackingFlag(Flag):
    """Tracking flags hold information how the user is being tracked"""
    REQUESTED = auto()
    """Tracking was explicitly requested by the user"""
    TRANSFER = auto()
    """Tracking was requested through transfer manager"""
    FRIEND = auto()
    """Tracking because the user is a friend"""


class TrackingState(Enum):
    UNTRACKED = "untracked"
    TRACKED = "tracked"
    RETRY_PENDING = "retry_pending"


class BlockingFlag(IntFlag):
    NONE = 0
    PRIVATE_MESSAGES = 1
    """Blocks private messages sent by the user"""
    ROOM_MESSAGES = 2
    """Blocks room messages sent by the user"""
    SEARCHES = 4
    """Blocks incoming searches from the user"""
    SHARES = 8
    """Blocks shares, directory listing requests"""
    INFO = 16
    UPLOADS = 32
    """Blocks all uploads and upload requests"""

    # Aliases
    IGNORE = PRIVATE_MESSAGES | ROOM_MESSAGES
    """Ignore all messaging"""
    ALL = IGNORE | SEARCHES | SHARES | INFO | UPLOADS
    """Fully blocks the user: includes messaging, shares requests, uploads"""


@dataclass(**slots_params)
class User:
    name: str

    # Status info
    status: UserStatus = UserStatus.UNKNOWN
    privileged: bool = False

    # User info
    description: Optional[str] = None
    picture: Optional[bytes] = None
    country: Optional[str] = None

    # User stats
    avg_speed: Optional[int] = None
    uploads: Optional[int] = None
    shared_file_count: Optional[int] = None
    shared_folder_count: Optional[int] = None

    # Upload status
    has_slots_free: Optional[bool] = None
    slots_free: Optional[int] = None
    upload_slots: Optional[int] = None
    queue_length: Optional[int] = None
    upload_permissions: UploadPermissions = UploadPermissions.UNKNOWN

    # Interests
    interests: set[str] = field(default_factory=set)
    hated_interests: set[str] = field(default_factory=set)

    def update_from_user_stats(self, user_stats: UserStats):
        self.avg_speed = user_stats.avg_speed
        self.shared_file_count = user_stats.shared_file_count
        self.shared_folder_count = user_stats.shared_folder_count
        self.uploads = user_stats.uploads

    def clear_all(self):
        """Clears all user data except user status and privileges info and
        country
        """
        self.clear(info=True, interests=True, stats=True, upload_info=True)

    def clear(
            self, info: bool = False, interests: bool = False,
            stats: bool = False, upload_info: bool = False):
        """Resets selected fields"""

        if info:
            self.picture = None
            self.description = None

        if interests:
            self.interests = set()
            self.hated_interests = set()

        if stats:
            self.avg_speed = None
            self.has_slots_free = None
            self.slots_free = None
            self.queue_length = None
            self.uploads = None

        if upload_info:
            self.upload_permissions = UploadPermissions.UNKNOWN
            self.upload_slots = None
            self.has_slots_free = None
            self.queue_length = None
            self.slots_free = None


@dataclass(**slots_params)
class ChatMessage:
    """Represents a private chat message"""
    id: int
    timestamp: int
    user: User
    message: str
    is_direct: bool

    def is_server_message(self) -> bool:
        return self.user.name == 'server'

    @property
    def is_admin(self) -> bool:
        """Only kept to keep backward compatibility. Use ``is_direct``

        This property does not actually represent whether the message was sent
        by an admin. Instead it represents whether the message was sent directly
        or was queued on the server before being sent (example, when user was
        offline and came back online)
        """
        return self.is_direct

    @is_admin.setter
    def is_admin(self, value: bool):
        self.is_direct = value



from __future__ import annotations
import asyncio
from collections import deque
from dataclasses import dataclass
from enum import Enum
import logging
import time
from typing import Optional

from .state import TransferState, TransferStateListener, VirginState

logger = logging.getLogger(__name__)


SPEED_LOG_INTERVAL = 0.1
SPEED_LOG_ENTRIES = 30


class FailReason:
    """Definition of reasons for which an transfer queue request or transfer
    request was rejected
    """
    CANCELLED = 'Cancelled'
    COMPLETE = 'Complete'
    QUEUED = 'Queued'
    FILE_NOT_SHARED = 'File not shared.'
    FILE_READ_ERROR = 'File read error.'


class AbortReason:
    REQUESTED = 'Requested'
    BLOCKED = 'Blocked'
    FILE_NOT_SHARED = 'File not shared'


class TransferDirection(Enum):
    UPLOAD = 0
    DOWNLOAD = 1


@dataclass(frozen=True, eq=True)
class TransferProgressSnapshot:
    """Represents the current progress of a transfer, used for reporting
    progress back to the user through a :class:`.TransferProgressEvent`
    """
    state: TransferState.State
    bytes_transfered: int
    """Amount of bytes transfered in bytes"""
    speed: float
    """Current transfer speed if the transfer is still downloading, average
    speed if the transfer has completed
    """
    start_time: Optional[float] = None
    complete_time: Optional[float] = None
    fail_reason: Optional[str] = None
    """Optional transfer fail reason"""
    abort_reason: Optional[str] = None
    """Optional transfer abort reason"""


class Transfer:
    """Class representing a transfer"""
    _UNPICKABLE_FIELDS = (
        '_speed_log',
        '_transfer_task',
        '_remotely_queue_task',
        '_state_lock',
        'progress_snapshot',
        'state_listeners'
    )

    def __init__(self, username: str, remote_path: str, direction: TransferDirection):
        self.state: TransferState = VirginState(self)

        self.direction: TransferDirection = direction
        """Determines whether this transfer is an upload or a download"""
        self.username: str = username
        """Username of the peer"""
        self.remote_path: str = remote_path
        """Remote path, this is the path that is shared between peers"""
        self.local_path: Optional[str] = None
        """Absolute path to the file on disk"""

        self.remotely_queued: bool = False
        """Indicates whether the transfer queue message was received by the peer.
        This only used for downloads
        """
        self.place_in_queue: Optional[int] = None
        self.fail_reason: Optional[str] = None
        """Reason for failure in case the transfer has failed"""
        self.abort_reason: Optional[str] = None
        """Reason for why transfer was aborted"""

        self.filesize: Optional[int] = None
        """Filesize in bytes"""

        self.bytes_transfered: int = 0
        """Amount of bytes transfered"""

        self.queue_attempts: int = 0
        self.last_queue_attempt: float = 0.0

        self.upload_request_attempts: int = 0
        self.last_upload_request_attempt: float = 0.0

        self.start_time: Optional[float] = None
        """Time at which the transfer was started. This is the time the transfer
        entered the DOWNLOADING or UPLOADING state
        """
        self.complete_time: Optional[float] = None
        """Time at which the transfer was completed. This is the time the
        transfer entered the COMPLETE, INCOMPLETE, ABORTED or FAILED state
        """

        self.progress_snapshot: TransferProgressSnapshot = self.take_progress_snapshot()
        """Snapshot of the transfer progress. Used internally to report progress
        at set intervals
        """
        self._speed_log: deque[tuple[float, int]] = deque(maxlen=SPEED_LOG_ENTRIES)
        self._remotely_queue_task: Optional[asyncio.Task] = None
        self._transfer_task: Optional[asyncio.Task] = None
        self._state_lock: asyncio.Lock = asyncio.Lock()
        self.state_listeners: list[TransferStateListener] = []

    def __setstate__(self, obj_state: dict):
        """Called when unpickling"""
        # Remove variables that are no longer used
        obj_state.pop('_offset', None)

        # Add variables that might not be in cache
        obj_state['state'] = TransferState.init_from_state(obj_state['state'], self)

        if 'abort_reason' not in obj_state:
            obj_state['abort_reason'] = None

        if obj_state['abort_reason'] is None:
            if obj_state['state'].VALUE == TransferState.ABORTED:
                obj_state['abort_reason'] = AbortReason.REQUESTED

        self.__dict__.update(obj_state)

        # Variables that are not stored in cache
        self._speed_log = deque(maxlen=SPEED_LOG_ENTRIES)
        self._remotely_queue_task = None
        self._transfer_task = None
        self._state_lock = asyncio.Lock()
        self.state_listeners = []
        self.progress_snapshot = self.take_progress_snapshot()

    def __getstate__(self):
        """Called when pickling, removes unpickable fields from the fields to
        store
        """
        obj_state = self.__dict__.copy()
        for unpickable_field in self._UNPICKABLE_FIELDS:
            if unpickable_field in obj_state:
                del obj_state[unpickable_field]

        obj_state['state'] = obj_state['state'].VALUE

        return obj_state

    def reset_local_vars(self):
        """Resets the local file variables"""
        self.local_path = None
        self.filesize = None

    def reset_progress_vars(self):
        self.bytes_transfered = 0
        self._offset = 0

    def reset_queue_vars(self):
        """Reset all variables when the """
        self.place_in_queue = None
        self.remotely_queued = False

        self.reset_queue_attempts()
        self.reset_upload_request_attempts()

    def reset_time_vars(self):
        """Clear all time related variables"""
        self.start_time = None
        self.complete_time = None
        self._speed_log = deque(maxlen=SPEED_LOG_ENTRIES)

    def set_start_time(self):
        """Set the start time, clear the complete time"""
        self.start_time = time.time()
        self.complete_time = None
        self._speed_log = deque(maxlen=SPEED_LOG_ENTRIES)

    def set_complete_time(self):
        """Set the complete time only if the start time has not been set"""
        if self.start_time is not None:
            self.complete_time = time.time()
            self._speed_log = deque(maxlen=SPEED_LOG_ENTRIES)

    def increase_queue_attempts(self):
        self.queue_attempts += 1
        self.last_queue_attempt = time.monotonic()

    def reset_queue_attempts(self):
        self.queue_attempts = 0
        self.last_queue_attempt = 0.0

    def increase_upload_request_attempt(self):
        self.upload_request_attempts += 1
        self.last_upload_request_attempt = time.monotonic()

    def reset_upload_request_attempts(self):
        self.upload_request_attempts = 0
        self.last_upload_request_attempt = 0.0

    async def transition(self, state: TransferState):
        """Transitions the state of the transfer and notifies the
        ``state_listeners``. This is an internal method and should not be used
        directly.
        """
        old_state = self.state
        logger.debug(
            "transitioning transfer state from %s to %s",
            old_state.VALUE.name, state.VALUE.name
        )
        self.state = state

        for listener in self.state_listeners:
            await listener.on_transfer_state_changed(
                self, old_state.VALUE, self.state.VALUE
            )

    def get_speed(self) -> float:
        """Retrieve the speed of the transfer

        :return: 0 if the transfer has not yet begun. The current speed if the
            transfer is ongoing. The transfer speed if the transfer was
            complete. Bytes per second
        """
        # Transfer hasn't begun
        if self.start_time is None:
            return 0.0

        # Transfer in progress
        if self.is_transferring():
            if len(self._speed_log) == 0:
                return 0.0

            current_time = time.monotonic()
            bytes_transfered = sum(bytes_sent for _, bytes_sent in self._speed_log)
            oldest_time = self._speed_log[0][0]

            transfer_duration = current_time - oldest_time
            if transfer_duration == 0.0:
                return 0.0

            return bytes_transfered / transfer_duration

        # Transfer complete
        if self.complete_time is None:
            transfer_duration = time.time() - self.start_time
        else:
            transfer_duration = self.complete_time - self.start_time

        if transfer_duration == 0.0:
            return 0.0

        return self.bytes_transfered / transfer_duration

    def add_speed_log_entry(self, bytes_transfered: int):
        """Logs a transfer speed entry. This is an internal method"""
        current_time = time.monotonic()
        # Create speed log entry
        if len(self._speed_log) == 0:
            self._speed_log.append((current_time, bytes_transfered))
        else:
            if current_time - self._speed_log[-1][0] < SPEED_LOG_INTERVAL:
                last_time_log, last_sent_bytes = self._speed_log.pop()
                self._speed_log.append((last_time_log, bytes_transfered + last_sent_bytes))
            else:
                self._speed_log.append((current_time, bytes_transfered))

    def is_upload(self) -> bool:
        return self.direction == TransferDirection.UPLOAD

    def is_download(self) -> bool:
        return self.direction == TransferDirection.DOWNLOAD

    def is_finalized(self) -> bool:
        """Return true if the transfer is in a finalized state"""
        return self.state.VALUE in (
            TransferState.COMPLETE,
            TransferState.ABORTED,
            TransferState.FAILED,
        )

    def is_processing(self) -> bool:
        """Return true if an attempt is being made to start transferring the
        file or the transfer is currently in progress.
        """
        return self.state.VALUE in (
            TransferState.DOWNLOADING,
            TransferState.UPLOADING,
            TransferState.INITIALIZING,
        )

    def is_transferring(self) -> bool:
        """Return true if the transfer is in progress"""
        return self.state.VALUE in (
            TransferState.DOWNLOADING,
            TransferState.UPLOADING,
        )

    def is_transfered(self) -> bool:
        return self.filesize == self.bytes_transfered

    def get_tasks(self) -> list[asyncio.Task]:
        tasks = []
        if self._remotely_queue_task is not None:
            tasks.append(self._remotely_queue_task)
        if self._transfer_task is not None:
            tasks.append(self._transfer_task)
        return tasks

    def cancel_tasks(self) -> list[asyncio.Task]:
        """Cancels all tasks for the transfer, this method returns the tasks
        which have been cancelled
        """
        tasks = []
        if self._remotely_queue_task is not None:
            tasks.append(self._remotely_queue_task)
            self._remotely_queue_task.cancel()

        if self._transfer_task is not None:
            tasks.append(self._transfer_task)
            self._transfer_task.cancel()

        return tasks

    def take_progress_snapshot(self) -> TransferProgressSnapshot:
        """Saves and returns a snapshot of the transfer progress"""
        snapshot = TransferProgressSnapshot(
            state=self.state.VALUE if isinstance(self.state, TransferState) else self.state,
            bytes_transfered=self.bytes_transfered,
            speed=self.get_speed(),
            start_time=self.start_time,
            complete_time=self.complete_time,
            fail_reason=self.fail_reason,
            abort_reason=self.abort_reason
        )
        self.progress_snapshot = snapshot
        return snapshot

    def _remotely_queue_task_complete(self, task: asyncio.Task):
        self._remotely_queue_task = None

    def _transfer_task_complete(self, task: asyncio.Task):
        self._transfer_task = None

    def _transfer_progress_callback(self, data: bytes):
        self.bytes_transfered += len(data)
        self.add_speed_log_entry(len(data))

    def __eq__(self, other: object):
        if not isinstance(other, Transfer):
            return NotImplemented

        other_vars = (other.remote_path, other.username, other.direction, )
        own_vars = (self.remote_path, self.username, self.direction, )
        return other_vars == own_vars

    def __str__(self) -> str:
        return (
            f"{self.__class__.__name__}(username={self.username!r}, "
            f"remote_path={self.remote_path!r}, direction={self.direction}, "
            f"local_path={self.local_path}, state={self.state})"
        )

    def __repr__(self) -> str:  # pragma: no cover
        return (
            f"{self.__class__.__name__}(username={self.username!r}, "
            f"remote_path={self.remote_path!r}, direction={self.direction})"
        )



from __future__ import annotations
import asyncio
from async_timeout import timeout as atimeout
from async_upnp_client.profiles.igd import PortMappingEntry
from dataclasses import dataclass
import enum
from functools import partial
from ipaddress import IPv4Address
import logging
from typing import (
    Any,
    Generic,
    Optional,
    Union,
    TypeVar,
    TYPE_CHECKING
)

from ..constants import (
    DEFAULT_LISTENING_HOST,
    PEER_INDIRECT_CONNECT_TIMEOUT,
)
from .connection import (
    Connection,
    ConnectionState,
    CloseReason,
    DataConnection,
    PeerConnectionState,
    ListeningConnection,
    PeerConnection,
    PeerConnectionType,
    ServerConnection,
)
from ..exceptions import (
    ConnectionFailedError,
    ConnectionReadError,
    ListeningConnectionFailedError,
    MessageDeserializationError,
    NetworkError,
    PeerConnectionError,
)
from ..events import (
    build_message_map,
    on_message,
    ConnectionStateChangedEvent,
    EventBus,
    PeerInitializedEvent,
    MessageReceivedEvent,
    SessionInitializedEvent,
    ServerReconnectedEvent,
)
from ..protocol.messages import (
    CannotConnect,
    ConnectToPeer,
    GetPeerAddress,
    MessageDataclass,
    PeerInit,
    PeerPierceFirewall,
    SetListenPort,
)
from . import upnp
from ..tasks import BackgroundTask
from .rate_limiter import RateLimiter
from ..utils import task_counter, ticket_generator

if TYPE_CHECKING:
    from ..settings import Settings


logger = logging.getLogger(__name__)


T = TypeVar('T', bound='MessageDataclass')
PeerConnectionT = TypeVar('PeerConnectionT', bound='PeerConnection')
ListeningConnections = tuple[Optional[ListeningConnection], Optional[ListeningConnection]]


class ListeningConnectionErrorMode(enum.Enum):
    """Error mode for listening connections. During initialization of the
    network the ``connect_listening_connections`` method will raise an error
    depending on the error mode
    """
    ALL = 'all'
    """Raise an exception if all listening connections failed to connect"""
    ANY = 'any'
    """Raise an exception if any listening connections failed to connect"""
    CLEAR = 'clear'
    """Raise an exception only if the non-obfuscated connection failed to
    connect
    """


class PeerConnectMode(enum.Enum):
    """Connection mode to use when connecting to peers"""
    FALLBACK = 'fallback'
    """Attempt direct connection first, fallback on indirect connection"""
    RACE = 'race'
    """Attempt both direct and indirect connection at the same time. First to
    connect will be used and the other will be disconnected
    """


class ExpectedResponse(asyncio.Future):
    """Future for an expected response message"""

    def __init__(
            self, connection_class: type[Union[PeerConnection, ServerConnection]],
            message_class: type[MessageDataclass],
            peer: Optional[str] = None, fields: Optional[dict[str, Any]] = None,
            loop: Optional[asyncio.AbstractEventLoop] = None):

        super().__init__(loop=loop)
        self.connection_class: type[Union[PeerConnection, ServerConnection]] = connection_class
        self.message_class: type[MessageDataclass] = message_class
        self.peer: Optional[str] = peer
        self.fields: dict[str, Any] = {} if fields is None else fields

    def matches(self, connection: DataConnection, response: MessageDataclass) -> bool:
        if connection.__class__ != self.connection_class:
            return False

        if response.__class__ != self.message_class:
            return False

        if self.peer is not None and isinstance(connection, PeerConnection):
            if connection.username != self.peer:
                return False

        for fname, expected_value in self.fields.items():
            if callable(expected_value):
                try:
                    actual_value = getattr(response, fname)
                except AttributeError:
                    return False
                else:
                    return expected_value(actual_value)

            if getattr(response, fname, None) != expected_value:
                return False

        return True


class PeerFuture(asyncio.Future, Generic[PeerConnectionT]):
    def __init__(
            self, ticket: int, username: str, typ: str,
            loop: Optional[asyncio.AbstractEventLoop] = None):

        super().__init__(loop=loop)
        self.ticket: int = ticket
        self.username: str = username
        self.typ: str = typ


@dataclass
class WatchdogContext:
    last_state: ConnectionState


class Network:

    def __init__(self, settings: Settings, event_bus: EventBus):
        self._settings: Settings = settings
        self._event_bus: EventBus = event_bus
        self._upnp = upnp.UPNP()
        self._ticket_generator = ticket_generator()
        self._expected_response_futures: list[ExpectedResponse] = []
        self._expected_connection_futures: dict[int, PeerFuture] = {}

        # List of connections
        self.server_connection: ServerConnection = self.create_server_connection()
        self.listening_connections: ListeningConnections = self.create_listening_connections()
        self.peer_connections: list[PeerConnection] = []

        self._MESSAGE_MAP = build_message_map(self)

        # Rate limiters
        self._upload_rate_limiter: RateLimiter = RateLimiter.create_limiter(
            self._settings.network.limits.upload_speed_kbps)
        self._download_rate_limiter: RateLimiter = RateLimiter.create_limiter(
            self._settings.network.limits.download_speed_kbps)

        # Debugging
        self._ip_overrides: dict[str, str] = self._settings.debug.ip_overrides

        # Operational
        self._create_peer_connection_tasks: list[asyncio.Task] = []
        self._log_connections_task: BackgroundTask = BackgroundTask(
            interval=10,
            task_coro=self._log_connections_job,
            preempt_wait=True,
            name='log-connections-task'
        )
        self._upnp_task: BackgroundTask = BackgroundTask(
            interval=self._settings.network.upnp.check_interval,
            task_coro=self._upnp_job,
            name='upnp-task'
        )
        self._connection_watchdog_task: BackgroundTask = BackgroundTask(
            interval=0.5,
            task_coro=self._server_connection_watchdog_job,
            name='server-connection-watchdog-task',
            context=WatchdogContext(self.server_connection.state)
        )

        self.register_listeners()

    def register_listeners(self):
        self._event_bus.register(
            SessionInitializedEvent, self._on_session_initialized)

    def create_server_connection(self) -> ServerConnection:
        return ServerConnection(
            self._settings.network.server.hostname,
            self._settings.network.server.port,
            self
        )

    def create_listening_connections(self) -> ListeningConnections:
        return (
            ListeningConnection(
                DEFAULT_LISTENING_HOST,
                self._settings.network.listening.port,
                self,
                obfuscated=False
            ) if self._settings.network.listening.port else None,
            ListeningConnection(
                DEFAULT_LISTENING_HOST,
                self._settings.network.listening.obfuscated_port,
                self,
                obfuscated=True
            ) if self._settings.network.listening.obfuscated_port else None
        )

    async def initialize(self):
        """Initializes the server and listening connections"""
        logger.info("initializing network")

        await self.connect_listening_ports()
        await self.connect_server()

        if self._settings.debug.log_connection_count:
            self._log_connections_task.start()

    async def connect_listening_ports(self):
        """This method will attempt to connect both listening ports (if
        configured) but will raise an exception depending on the settings
        ``network.listening.error_mode`` setting. All other listening connections
        will be disconnected before raising the error

        :raise ListeningConnectionFailedError: if an error occurred connecting
            the listening ports
        """
        error_mode = self._settings.network.listening.error_mode
        results = await asyncio.gather(
            *[
                listening_connection.connect()
                for listening_connection in self.listening_connections
                if listening_connection
            ],
            return_exceptions=True
        )

        if error_mode == ListeningConnectionErrorMode.ALL:
            if all(isinstance(res, ConnectionFailedError) for res in results):
                await self.disconnect_listening_ports()
                raise ListeningConnectionFailedError("failed to open any listening ports")

        elif error_mode == ListeningConnectionErrorMode.ANY:
            if any(isinstance(res, ConnectionFailedError) for res in results):
                await self.disconnect_listening_ports()
                raise ListeningConnectionFailedError("one or more listening ports failed to connect")

        elif error_mode == ListeningConnectionErrorMode.CLEAR:
            if not self.listening_connections[0] or self.listening_connections[0].state != ConnectionState.CONNECTED:
                await self.disconnect_listening_ports()
                raise ListeningConnectionFailedError("failed to connect non-obfuscated listening port")

    async def disconnect_listening_ports(self):
        """Disconnects all listening ports"""
        await asyncio.gather(
            *[
                listening_connection.disconnect(CloseReason.REQUESTED)
                for listening_connection in self.listening_connections
                if listening_connection
            ],
            return_exceptions=True
        )

    async def connect_server(self):
        await self.server_connection.connect()

    async def disconnect_server(self):
        await self.server_connection.disconnect(CloseReason.REQUESTED)

    async def disconnect(self):
        """Cancels all connection creation tasks and disconnects all current
        open connections
        """
        self._cancel_all_tasks()

        connections: list[Connection] = [self.server_connection]
        connections.extend(self.peer_connections)
        connections.extend(conn for conn in self.listening_connections if conn)

        logger.info("waiting for network disconnect : %d connections", len(connections))
        await asyncio.gather(
            *[conn.disconnect(CloseReason.REQUESTED) for conn in connections],
            return_exceptions=True
        )
        logger.info("network disconnected")

    def get_listening_ports(self) -> tuple[int, int]:
        """Gets the currently connected listening ports

        :return: ``tuple`` with 2 elements: the non-obfuscated- and obfuscated
            ports. If either is not connected or not configured ``0`` will be
            returned for this port
        """
        def get_port(conn: Optional[ListeningConnection]):
            if conn and conn.state == ConnectionState.CONNECTED:
                return conn.port
            return 0

        nonobf_port = get_port(self.listening_connections[0])
        obf_port = get_port(self.listening_connections[1])
        return nonobf_port, obf_port

    async def advertise_listening_ports(self):
        """Notifies the server of our current listening ports"""
        port, obfuscated_port = self.get_listening_ports()

        await self.send_server_messages(
            SetListenPort.Request(
                port,
                obfuscated_port_amount=1 if obfuscated_port else 0,
                obfuscated_port=obfuscated_port
            )
        )

    # Settings listeners
    def load_speed_limits(self):
        """(Re)loads the speed limits from the settings"""
        self.set_download_speed_limit(
            self._settings.network.limits.download_speed_kbps)
        self.set_upload_speed_limit(
            self._settings.network.limits.upload_speed_kbps)

    def set_upload_speed_limit(self, limit_kbps: int):
        """Modifies the upload speed limit. Passing 0 will set the upload speed
        to unlimited

        :param limit_kbps: the new upload limit
        """
        new_limiter = RateLimiter.create_limiter(limit_kbps)
        if self._upload_rate_limiter is not None:
            new_limiter.copy_tokens(self._upload_rate_limiter)
        self._upload_rate_limiter = new_limiter

        for conn in self.peer_connections:
            conn.upload_rate_limiter = self._upload_rate_limiter

    def set_download_speed_limit(self, limit_kbps: int):
        """Modifies the download speed limit. Passing 0 will set the download
        speed to unlimited

        :param limit_kbps: the new download limit
        """
        new_limiter = RateLimiter.create_limiter(limit_kbps)
        if self._download_rate_limiter is not None:
            new_limiter.copy_tokens(self._download_rate_limiter)
        self._download_rate_limiter = new_limiter

        for conn in self.peer_connections:
            conn.download_rate_limiter = self._download_rate_limiter

    async def _server_connection_watchdog_job(self, context: WatchdogContext):
        """Reconnects to the server if it is closed. This should be started as a
        task and should be cancelled upon request.
        """
        timeout = self._settings.network.server.reconnect.timeout
        if self.server_connection.state == ConnectionState.CLOSED:

            if not self._settings.credentials.are_configured():
                if self.server_connection.state != context.last_state:
                    logger.warning(
                        "credentials are not correctly configured in the "
                        "settings, will not attempt to reconnect"
                    )
                return

            else:
                logger.info("will attempt to reconnect to server in %d seconds", timeout)
                await asyncio.sleep(timeout)
                try:
                    await self.connect_server()
                except ConnectionFailedError:
                    logger.warning("failed to reconnect to server")
                else:
                    await self._event_bus.emit(
                        ServerReconnectedEvent())

        context.last_state = self.server_connection.state

    async def _log_connections_job(self):
        count = len(self.peer_connections)
        logger.info(
            "currently %d peer connections (%d tasks)",
            count, len(self._create_peer_connection_tasks)
        )

    async def _upnp_job(self):
        """This job ensures the UPnP port mappings are created and maintained.

        The interval that the mapping is checked is configured through the
        ``network.upnp.check_interval`` setting. This setting represents the max
        interval, the check interval could be shorter if the job detects that
        the lease duration of a port is about to expire
        """
        def filter_mapped_port(mapped_ports: list[PortMappingEntry], internal_ip: str, port: int) -> PortMappingEntry:
            expected_values = (True, 'TCP', IPv4Address(internal_ip), port)
            for mapped_port in mapped_ports:
                values = (
                    mapped_port.enabled,
                    mapped_port.protocol,
                    mapped_port.internal_client,
                    mapped_port.internal_port
                )
                if values == expected_values:
                    return mapped_port

            raise ValueError(
                f"port {internal_ip}:{port} was not mapped on device {device.name}")

        ports = [port for port in self.get_listening_ports() if port]
        ip_address = self.server_connection.get_connecting_ip()

        devices = await self._upnp.search_igd_devices(
            ip_address, timeout=self._settings.network.upnp.search_timeout
        )
        logger.info("found %d devices for UPnP port mapping", len(devices))
        lease_expirations = []

        for device in devices:
            device_mapped_ports = await self._upnp.get_mapped_ports(device)
            for port in ports:
                try:
                    mapping = filter_mapped_port(
                        device_mapped_ports, ip_address, port)

                except ValueError:
                    try:
                        logger.info(
                            "UPnP: mapping port %s:%d on device %s",
                            ip_address, port, device.name
                        )
                        await self._upnp.map_port(
                            device, ip_address, port,
                            lease_duration=self._settings.network.upnp.lease_duration
                        )
                        logger.info(
                            "UPnP: mapped port %s:%d on device %s",
                            ip_address, port, device.name
                        )

                    except Exception as exc:
                        logger.warning(
                            "UPnP: failed to map port %s:%d on device %s",
                            ip_address, port, device.name,
                            exc_info=exc
                        )

                except Exception as exc:
                    logger.warning(
                        "UPnP: failed to get port mapping %s:%d on device %s",
                        ip_address, port, device.name,
                        exc_info=exc
                    )
                else:
                    if lease_duration := mapping.lease_duration:
                        lease_expirations.append(lease_duration.total_seconds())

        next_check = min(
            lease_expirations + [self._settings.network.upnp.check_interval])
        logger.info("UPnP: rechecking port mapping in %d seconds", next_check)
        return next_check

    async def start_server_connection_watchdog(self):
        """Starts the server connection watchdog if it is not already running"""
        self._connection_watchdog_task.start()

    def stop_server_connection_watchdog(self):
        """Stops the server connection watchdog if it is running"""
        self._connection_watchdog_task.cancel()

    async def start_upnp_job(self):
        self._upnp_task.start()

    def stop_upnp_job(self):
        self._upnp_task.cancel()

    def select_port(self, port, obfuscated_port) -> tuple[int, bool]:
        """Selects the port used for making a connection. This attempts to take
        into account the ``network.peer.obfuscate`` parameter however falls back
        on using whatever port is available if necessary

        :param port: available non-obfuscated port
        :param obfuscated_port: available obfuscated port
        :return: a tuple with the port number and whether the port is obfuscated
        """
        prefer_obfuscated = self._settings.network.peer.obfuscate
        if port and obfuscated_port:
            if prefer_obfuscated:
                return obfuscated_port, True
            else:
                return port, False
        elif port:
            return port, False
        else:
            return obfuscated_port, True

    async def create_peer_connection(
            self, username: str, typ: str, ip: Optional[str] = None, port: Optional[int] = None,
            obfuscate: bool = False) -> PeerConnection:
        """Creates a new peer connection to the given ``username`` and
        connection type.

        Optionally this method takes ``ip``, ``port`` and ``obfuscate``
        parameters. If not given it will be requested to the server
        (GetPeerAddress) before attempting connection

        :param username: username of the peer
        :param typ: connection type ('P', 'D', 'F')
        :param ip: optional ip address of the peer
        :param port: optional port of the peer
        :param obfuscate: whether to obfuscate the connection. Should be used in
            combination with the ``ip`` and ``port``
        :return: an object of `PeerConnection`
        :raise PeerConnectionError: when establishing a peer connection has
            failed
        """
        ticket = next(self._ticket_generator)

        logger.debug("creating new peer connection to %s (type=%s, ticket=%d)", username, typ, ticket)

        if self._settings.network.peer.connect_mode == PeerConnectMode.RACE:
            connection = await self._create_peer_connection_race(
                ticket, username, typ, ip=ip, port=port, obfuscate=obfuscate)

        else:
            connection = await self._create_peer_connection_fallback(
                ticket, username, typ, ip=ip, port=port, obfuscate=obfuscate)

        return connection

    async def _create_peer_connection_fallback(
            self, ticket: int, username: str, typ: str,
            ip: Optional[str] = None, port: Optional[int] = None, obfuscate: bool = False):

        try:
            connection = await self._make_direct_connection(
                ticket, username, typ, ip, port, obfuscate
            )

        except NetworkError as exc:
            logger.debug(
                "direct connection (%s) to peer failed : %s : %r",
                typ, username, exc
            )

            try:
                connection = await self._make_indirect_connection(
                    ticket, username, typ
                )

            except NetworkError as exc:
                logger.debug("indirect connection to peer failed : %s : %r", username, exc)
                raise PeerConnectionError(
                    f"failed to connect to peer {username} ({typ=}, {ticket=})")

        return connection

    async def _create_peer_connection_race(
            self, ticket: int, username: str, typ: str,
            ip: Optional[str] = None, port: Optional[int] = None, obfuscate: bool = False) -> PeerConnection:

        direct_task = asyncio.create_task(
            self._make_direct_connection(
                ticket, username, typ, ip=ip, port=port, obfuscate=obfuscate
            ),
            name=f"direct-connect-{username}-{typ}-{ticket}"
        )
        indirect_task = asyncio.create_task(
            self._make_indirect_connection(ticket, username, typ),
            name=f"indirect-connect-{username}-{typ}-{ticket}"
        )

        pending = {direct_task, indirect_task}
        while pending:
            done, pending = await asyncio.wait(pending, return_when=asyncio.FIRST_COMPLETED)

            connections = []
            for done_task in done:
                try:
                    connections.append(done_task.result())
                except Exception:
                    pass

            if connections:

                if pending:
                    for pending_task in pending:
                        logger.debug("cancelling connect task : %s", pending_task.get_name())
                        pending_task.cancel()
                    await asyncio.gather(*pending, return_exceptions=True)

                if len(connections) > 1:
                    await connections[1].disconnect(CloseReason.REQUESTED)

                return connections[0]

        raise PeerConnectionError(
            f"failed to connect to peer {username} ({typ=}, {ticket=})")

    async def _get_peer_address(self, username: str) -> tuple[str, int, int]:
        """Requests the peer address for the given ``username`` from the server.

        :raise PeerConnectionError: if no IP address or no valid ports were
            returned
        """
        await self.server_connection.send_message(GetPeerAddress.Request(username))
        _, response = await self.create_server_response_future(
            GetPeerAddress.Response,
            fields={
                'username': username
            }
        )

        if response.ip == '0.0.0.0':
            logger.warning("GetPeerAddress : no address returned for username : %s", username)
            raise PeerConnectionError(f"no address for user : {username}")

        elif not response.port and not response.obfuscated_port:
            logger.warning("GetPeerAddress : no valid ports found for user : %s", username)
            raise PeerConnectionError(f"no valid ports for user : {username}")

        return response.ip, response.port, response.obfuscated_port

    async def get_peer_connection(self, username: str, typ: str = PeerConnectionType.PEER) -> PeerConnection:
        """Gets a peer connection for the given ``username``. It will first try
        to re-use an existing connection, otherwise it will create a new
        connection

        :param username: username of the peer to connect to
        :param typ: Type of connection to create or reuse
        :return: A :class:`.PeerConnection` instance
        """
        connections = self.get_active_peer_connections(username, typ)
        if connections:
            return connections[0]

        return await self.create_peer_connection(username, typ)

    def _remove_response_future(self, expected_response: ExpectedResponse):
        if expected_response in self._expected_response_futures:
            self._expected_response_futures.remove(expected_response)

    def _remove_connection_future(self, ticket: int, connection_future: asyncio.Future):
        logger.debug("removing expected incoming connection with ticket : %d", ticket)
        self._expected_connection_futures.pop(ticket)

    def create_server_response_future(
            self, message_class: type[T], fields: Optional[dict[str, Any]] = None) -> ExpectedResponse:
        """Creates a future for a server message to arrive, the message must
        match the ``message_class`` and fields defined in the keyword arguments.

        The future will be stored by the current class and will be removed once
        the future is complete or cancelled.

        :param message_class: The expected :class:`.MessageDataClass`
        :return: :class:`.ExpectedResponse` object
        """
        fields = fields or {}
        future = ExpectedResponse(
            ServerConnection,
            message_class=message_class,
            fields=fields
        )
        self._expected_response_futures.append(future)
        future.add_done_callback(self._remove_response_future)
        return future

    def register_response_future(self, expected_response: ExpectedResponse):
        """Registers an expected response future"""
        self._expected_response_futures.append(expected_response)
        expected_response.add_done_callback(self._remove_response_future)

    async def wait_for_server_message(
            self, message_class: type[T], fields: Optional[dict[str, Any]] = None, timeout: float = 10) -> T:
        """Waits for a message from the server

        :param message_class: Class of the expected server message
        :param fields: Optional matchers for the message fields
        :return: The :class:`.MessageData` object corresponding to the response
        :raise: TimeoutError
        """
        future = self.create_server_response_future(
            message_class=message_class,
            fields=fields
        )
        try:
            async with atimeout(timeout):
                _, response = await future
        except TimeoutError as exc:
            future.set_exception(exc)
            raise

        return response

    def create_peer_response_future(
            self, peer: str, message_class: type[T], fields: Optional[dict[str, Any]] = None) -> ExpectedResponse:
        """Creates a future for a peer message to arrive, the message must match
        the ``message_class`` and fields defined in the keyword arguments and
        must be coming from a connection by ``peer``.

        The future will be stored by the current class and will be removed once
        the future is complete or cancelled.

        :param peer: name of the peer for which the message should be received
        :return: :class:`.ExpectedResponse` object
        """
        fields = fields or {}
        future = ExpectedResponse(
            PeerConnection,
            message_class=message_class,
            peer=peer,
            fields=fields
        )
        self._expected_response_futures.append(future)
        future.add_done_callback(self._remove_response_future)
        return future

    async def wait_for_peer_message(
            self, peer: str, message_class: type[T], fields: Optional[dict[str, Any]] = None, timeout: float = 60) -> T:

        future = self.create_peer_response_future(
            peer=peer,
            message_class=message_class,
            fields=fields
        )
        try:
            async with atimeout(timeout):
                _, response = await future
        except TimeoutError as exc:
            future.set_exception(exc)
            raise

        return response

    def get_peer_connections(self, username: str, typ: str) -> list[PeerConnection]:
        """Returns all connections for peer with given username and peer
        connection types.

        :param username: username of the peer
        :param typ: type of the connection (from :class:`.PeerConnectionType`)
        :return: list of connections for that matches the parameters
        """
        return [
            connection for connection in self.peer_connections
            if connection.username == username and connection.connection_type == typ
        ]

    def get_active_peer_connections(
            self, username: str, typ: str) -> list[PeerConnection]:
        """Return a list of currently active messaging connections for given
        peer connection type.

        :param username: username for which to get the active peer
        :param typ: peer connection type
        :return: list of PeerConnection instances
        """
        return [
            conn
            for conn in self.get_peer_connections(username, typ)
            if conn.state == ConnectionState.CONNECTED and conn.connection_state == PeerConnectionState.ESTABLISHED
        ]

    def remove_peer_connection(self, connection: PeerConnection):
        if connection in self.peer_connections:
            self.peer_connections.remove(connection)

    # Server related

    @on_message(ConnectToPeer.Response)
    async def _on_connect_to_peer(self, message: ConnectToPeer.Response, connection: ServerConnection):
        """Handles a :class:`ConnectToPeer` request coming from another peer"""
        task = asyncio.create_task(
            self._handle_connect_to_peer(message),
            name=f'connect-to-peer-{task_counter()}'
        )
        task.add_done_callback(
            partial(self._handle_connect_to_peer_callback, message)
        )
        self._create_peer_connection_tasks.append(task)

    async def _make_direct_connection(
            self, ticket: int, username: str, typ: str, ip: Optional[str] = None,
            port: Optional[int] = None, obfuscate: bool = False) -> PeerConnection:
        """Attempts to make a direct connection to the peer and send a
        :class:`.PeerInit` message. This will be the first step in case we are
        the one initiating the connection
        """
        # Request peer address if ip and port are not given
        if ip is None or port is None:
            ip, clear_port, obfuscated_port = await self._get_peer_address(username)

            port, obfuscate = self.select_port(clear_port, obfuscated_port)

        # Override IP address if requested
        ip = self._ip_overrides.get(username, ip)

        logger.debug(
            "attempting to connect directly to peer : %r %s:%d",
            username, ip, port
        )
        connection = PeerConnection(
            ip, port, self,
            connection_type=typ,
            username=username,
            obfuscated=obfuscate
        )
        self.peer_connections.append(connection)

        await connection.connect()
        await connection.send_message(
            PeerInit.Request(
                self._settings.credentials.username,
                typ,
                ticket
            )
        )

        self._finalize_peer_connection(connection)

        await self._event_bus.emit(
            PeerInitializedEvent(connection, requested=True))

        return connection

    async def _make_indirect_connection(
            self, ticket: int, username: str, typ: str) -> PeerConnection:
        """Attempts to make an indirect connection by sending a
        :class:`.ConnectToPeer` message to the server. The method will wait for
        an incoming connection with a :class:`.PeerPierceFirewall` message
        containing the provided ticket or a :class:`.CannotConnect` message from
        the server.

        :return: :class:`.PeerConnection` object in case of success
        :raise PeerConnectionError: in case timeout was reached or we received a
            :class:`.CannotConnect` message from the server containing the
            ``ticket``
        """
        # Wait for either a established connection with PeerPierceFirewall or a
        # CannotConnect from the server
        expected_connection_future: PeerFuture = PeerFuture(ticket, username, typ)
        expected_connection_future.add_done_callback(
            partial(self._remove_connection_future, ticket)
        )
        self._expected_connection_futures[ticket] = expected_connection_future

        cannot_connect_future = self.create_server_response_future(
            CannotConnect.Response,
            fields={
                'ticket': ticket
            }
        )

        # Send the connect to peer message
        await self.server_connection.send_message(
            ConnectToPeer.Request(ticket, username, typ))

        futures = (expected_connection_future, cannot_connect_future)
        done, pending = await asyncio.wait(
            futures,
            timeout=PEER_INDIRECT_CONNECT_TIMEOUT,
            return_when=asyncio.FIRST_COMPLETED
        )

        # Whatever happens here, we can cancel all pending futures
        [fut.cancel() for fut in pending]

        # `done` will be empty in case of timeout
        if not done:
            raise PeerConnectionError(
                f"indirect connection timed out ({username=}, {ticket=})")

        completed_future = done.pop()

        if completed_future == cannot_connect_future:
            raise PeerConnectionError(
                f"indirect connection failed ({username=}, {ticket=})")

        return completed_future.result()

    async def _handle_connect_to_peer(self, message: ConnectToPeer.Response):
        """Handles an indirect connection request received from the server.

        A task is created for this coroutine when a :class:`.ConnectToPeer`
        message is received.

        A :class:`.PeerInitializedEvent` will be emitted if the connection has
        been successfully established

        :param message: received :class:`.ConnectToPeer` message
        """
        ip = self._ip_overrides.get(message.username, message.ip)
        port, obfuscate = self.select_port(message.port, message.obfuscated_port)

        peer_connection = PeerConnection(
            ip, port, self,
            connection_type=message.typ,
            obfuscated=obfuscate,
            username=message.username
        )
        self.peer_connections.append(peer_connection)

        try:
            await peer_connection.connect()
            await peer_connection.send_message(
                PeerPierceFirewall.Request(message.ticket)
            )

        except NetworkError:
            await self.server_connection.send_message(
                CannotConnect.Request(
                    ticket=message.ticket,
                    username=message.username
                )
            )
            raise PeerConnectionError("failed connect on user request")

        self._finalize_peer_connection(peer_connection)

        await self._event_bus.emit(
            PeerInitializedEvent(peer_connection, requested=False))

    def _finalize_peer_connection(self, connection: PeerConnection):
        if connection.connection_type == PeerConnectionType.FILE:
            connection.set_connection_state(PeerConnectionState.NEGOTIATING_TRANSFER)
            connection.download_rate_limiter = self._download_rate_limiter
            connection.upload_rate_limiter = self._upload_rate_limiter

        else:
            connection.set_connection_state(PeerConnectionState.ESTABLISHED)

    async def send_peer_messages(
            self, username: str, *messages: Union[bytes, MessageDataclass],
            raise_on_error: bool = True):
        """Sends a list of messages to the peer with given ``username``. This
        uses ``get_peer_connection`` and will attempt to re-use a connection or
        create a new peer (P) connection

        :param username: Peer username to send the messages to
        :param messages: List of messages to send
        :param raise_on_error: When ``True`` an exception is raised when a
            message failed to send, if ``False`` a list of tuples with the
            result for each message will be returned
        :return: a list of tuples containing the sent messages and the result
            of the sent messages. If ``None`` is returned for a message it was
            successfully sent, otherwise the result will contain the exception.
            `None` if the ``raise_on_error`` is ``True``
        """
        connection = await self.get_peer_connection(username)
        results = await asyncio.gather(
            *[connection.send_message(message) for message in messages],
            return_exceptions=not raise_on_error
        )
        if not raise_on_error:
            return list(zip(messages, results))

    def queue_server_messages(self, *messages: Union[bytes, MessageDataclass]) -> list[asyncio.Task]:
        """Queues server messages

        :param messages: list of messages to queue
        """
        return self.server_connection.queue_messages(*messages)

    async def send_server_messages(
            self, *messages: Union[bytes, MessageDataclass],
            raise_on_error: bool = True):
        """Sends a list of messages to the server

        :param messages: List of messages to send
        :param raise_on_error: When ``True`` an exception is raised when a
            message failed to send, if ``False`` a list of tuples with the
            result for each message will be returned
        :return: a list of tuples containing the sent messages and the result
            of the sent messages. If `None` is returned for a message it was
            successfully sent, otherwise the result will contain the exception.
            ``None`` if the ``raise_on_error`` is ``True``
        """
        results = await asyncio.gather(
            *[self.server_connection.send_message(message) for message in messages],
            return_exceptions=not raise_on_error
        )
        if not raise_on_error:
            return list(zip(messages, results))

    # Connection state changes
    async def on_state_changed(
            self, state: ConnectionState, connection: Connection,
            close_reason: CloseReason = CloseReason.UNKNOWN):
        """Called when the state of a connection changes. This method calls 3
        private method based on the type of ``connection`` that was passed

        :param state: the new state the connection has received
        :param connection: the connection for which the state changed
        :param close_reason: in case ConnectionState.CLOSED is passed a reason
            will be given as well
        """
        if isinstance(connection, ServerConnection):
            await self._on_server_connection_state_changed(state, connection, close_reason=close_reason)

        elif isinstance(connection, PeerConnection):
            await self._on_peer_connection_state_changed(state, connection, close_reason=close_reason)

        await self._event_bus.emit(
            ConnectionStateChangedEvent(connection, state, close_reason)
        )

    async def _on_server_connection_state_changed(
            self, state: ConnectionState, connection: ServerConnection,
            close_reason: CloseReason = CloseReason.UNKNOWN):

        if state == ConnectionState.CONNECTED:
            # For registering with UPNP we need to know our own IP first, we can
            # get this from the server connection but we first need to be
            # fully connected to it before we can request a valid IP
            if self._settings.network.upnp.enabled:
                await self.start_upnp_job()

            # Register server connection watchdog
            if self._settings.network.server.reconnect.auto:
                await self.start_server_connection_watchdog()

        elif state == ConnectionState.CLOSING:

            self.stop_upnp_job()
            # When `disconnect` is called on the connection it will always first
            # go into the CLOSING state. The watchdog will only attempt to
            # reconnect if the server is in CLOSED state. So the code needs to
            # make a decision here whether we want to reconnect or not before it
            # goes into CLOSED state.
            # Cancel the watchdog only if we are closing up on request or the
            # server sends an EOF. The server sends an EOF as soon as you are
            # connected if your IP address is banned, this happens when you make
            # too many connections in a short period of time. Making any more
            # connections will only extend your ban period. Possibly the server
            # will also EOF when the login message is not sent after a certain
            # period of time
            # An EOF will also be sent in case you are kicked due to the same
            # login being used in another location
            if close_reason == CloseReason.REQUESTED:
                self.stop_server_connection_watchdog()

            elif close_reason == CloseReason.EOF:
                logger.warning(
                    "server closed connection, will not attempt to reconnect")
                self.stop_server_connection_watchdog()

    async def _on_peer_connection_state_changed(
            self, state: ConnectionState, connection: PeerConnection,
            close_reason: CloseReason = CloseReason.UNKNOWN):

        if state == ConnectionState.CLOSED:
            self.remove_peer_connection(connection)

    # Peer related
    async def on_peer_accepted(self, connection: PeerConnection):
        """Called when a connection has been accepted on one of the listening
        connections. This method will wait for the peer initialization message:
        either :class:`.PeerInit` or :class:`.PeerPierceFirewall`.

        In case of :class:`.PeerInit` the method will perform the initialization
        of the connection.

        In case of :class:`.PeerPierceFirewall` we look for an associated future
        object for the ticket provided in the message; there should be a task
        waiting for it to be completed (see ``_make_indirect_connection``). Full
        initialization of the connection should be done in that method.

        In any other case we disconnect the connection.

        :param connection: the accepted :class:`.PeerConnection`
        """
        self.peer_connections.append(connection)

        try:
            peer_init_message = await connection.receive_message_object()
            if not peer_init_message:
                # EOF reached before receiving a message. Connection should've
                # been automatically closed
                return
        except ConnectionReadError:
            # Some read error. Connection should've been automatically closed
            return
        except MessageDeserializationError:
            # Couldn't deserialize the first message. Assume something is broken
            # and disconnect
            await connection.disconnect(CloseReason.READ_ERROR)
            return

        if isinstance(peer_init_message, PeerInit.Request):
            connection.username = peer_init_message.username
            connection.connection_type = peer_init_message.typ
            self._finalize_peer_connection(connection)

            await self._event_bus.emit(
                PeerInitializedEvent(connection, requested=False))

        elif isinstance(peer_init_message, PeerPierceFirewall.Request):
            ticket = peer_init_message.ticket
            try:
                connection_future = self._expected_connection_futures[ticket]
            except KeyError:
                logger.warning(
                    "%s:%d : unknown pierce firewall ticket : %d",
                    connection.hostname, connection.port, ticket
                )
                await connection.disconnect(CloseReason.REQUESTED)
            else:
                connection.username = connection_future.username
                connection.connection_type = connection_future.typ
                self._finalize_peer_connection(connection)

                await self._event_bus.emit(
                    PeerInitializedEvent(connection, requested=True))

                connection_future.set_result(connection)

        else:
            logger.warning(
                "%s:%d : unknown peer init message : %s",
                connection.hostname, connection.port, peer_init_message
            )
            await connection.disconnect(CloseReason.REQUESTED)

    async def on_message_received(self, message: MessageDataclass, connection: DataConnection):
        """Method called by ``connection`` instances when a message is received

        :param message: Received message instance
        :param connection: Connection on which the message was received
        """
        # Call the message callbacks for this object
        if message.__class__ in self._MESSAGE_MAP:
            await self._MESSAGE_MAP[message.__class__](message, connection)

        # Emit on the message bus. Event handling has priority over completing
        # the future responses
        await self._event_bus.emit(MessageReceivedEvent(message, connection))

        # Complete expected response futures
        for expected_response in self._expected_response_futures:
            if expected_response.matches(connection, message):
                expected_response.set_result((connection, message, ))

    async def _on_session_initialized(self, event: SessionInitializedEvent):
        await self.advertise_listening_ports()

    # Task callbacks
    def _handle_connect_to_peer_callback(self, message: ConnectToPeer.Response, task: asyncio.Task):
        try:
            task.result()
        except asyncio.CancelledError:
            logger.debug("cancelled ConnectToPeer request : %d", message)
        except Exception as exc:
            logger.warning("failed to fulfill ConnectToPeer request : %s : %r", message, exc)
        else:
            logger.info("successfully fulfilled ConnectToPeer request : %s", message)
        finally:
            self._create_peer_connection_tasks.remove(task)

    def _cancel_all_tasks(self):
        for task in self._create_peer_connection_tasks:
            task.cancel()

        self._log_connections_task.cancel()
        self._upnp_task.cancel()



import asyncio
from collections import deque
from functools import partial
import logging
from typing import Optional, Union

from ..base_manager import BaseManager
from ..constants import DEFAULT_WISHLIST_INTERVAL
from ..events import (
    on_message,
    build_message_map,
    ConnectionStateChangedEvent,
    EventBus,
    MessageReceivedEvent,
    SearchRequestReceivedEvent,
    SearchRequestRemovedEvent,
    SearchRequestSentEvent,
    SearchResultEvent,
    SessionDestroyedEvent,
    SessionInitializedEvent,
)
from ..protocol.messages import (
    DistributedSearchRequest,
    DistributedServerSearchRequest,
    ExcludedSearchPhrases,
    FileSearch,
    PeerSearchReply,
    RoomSearch,
    SearchInactivityTimeout,
    ServerSearchRequest,
    UserSearch,
    WishlistInterval,
    WishlistSearch,
)
from ..network.connection import (
    CloseReason,
    ConnectionState,
    PeerConnection,
    ServerConnection,
)
from ..network.network import Network
from ..room.model import Room
from ..settings import Settings
from ..shares.manager import SharesManager
from ..shares.utils import convert_items_to_file_data
from ..session import Session
from ..tasks import BackgroundTask, Timer
from ..transfer.interface import UploadInfoProvider
from ..user.model import BlockingFlag
from ..utils import task_counter, ticket_generator
from .model import ReceivedSearch, SearchResult, SearchRequest, SearchType


logger = logging.getLogger(__name__)


class SearchManager(BaseManager):
    """Handler for searches requests"""

    def __init__(
            self, settings: Settings, event_bus: EventBus,
            shares_manager: SharesManager, upload_info_provider: UploadInfoProvider,
            network: Network):

        self._settings: Settings = settings
        self._event_bus: EventBus = event_bus
        self._network: Network = network
        self._shares_manager: SharesManager = shares_manager
        self._upload_info_provider: UploadInfoProvider = upload_info_provider

        self._ticket_generator = ticket_generator()
        self._session: Optional[Session] = None

        self.received_searches: deque[ReceivedSearch] = deque(
            list(), maxlen=self._settings.searches.receive.store_amount)
        self.requests: dict[int, SearchRequest] = {}

        # Server variables
        self.search_inactivity_timeout: Optional[int] = None
        self.wishlist_interval: Optional[int] = None
        self.excluded_search_phrases: list[str] = []

        self.register_listeners()

        self._MESSAGE_MAP = build_message_map(self)

        self._search_reply_tasks: list[asyncio.Task] = []
        self._wishlist_task: BackgroundTask = BackgroundTask(
            interval=DEFAULT_WISHLIST_INTERVAL,
            task_coro=self._wishlist_job,
            name='wishlist-task'
        )

    def register_listeners(self):
        self._event_bus.register(
            ConnectionStateChangedEvent, self._on_state_changed)
        self._event_bus.register(
            MessageReceivedEvent, self._on_message_received)
        self._event_bus.register(
            SessionInitializedEvent, self._on_session_initialized)
        self._event_bus.register(
            SessionDestroyedEvent, self._on_session_destroyed)

    def remove_request(self, request: Union[SearchRequest, int]):
        """Removes the search request from the client. Incoming results after
        the request has been removed will be ignored

        :param request: :class:`.SearchRequest` object or ticket number to
            remove
        """
        ticket = request if isinstance(request, int) else request.ticket
        self.requests.pop(ticket)

    async def search(self, query: str) -> SearchRequest:
        """Performs a global search. The results generated by this query will
        stored in the returned object or can be listened to through the
        :class:`.SearchResultEvent` event

        :param query: The search query
        :return: An object containing the search request details and results
        """
        ticket = next(self._ticket_generator)

        await self._network.send_server_messages(
            FileSearch.Request(ticket, query)
        )
        request = SearchRequest(
            ticket=ticket,
            query=query,
            search_type=SearchType.NETWORK
        )
        await self._attach_request_timer_and_emit(request)

        return request

    async def search_room(self, room: Union[str, Room], query: str) -> SearchRequest:
        """Performs a search request on the specific room. The results generated
        by this query will stored in the returned object or can be listened to
        through the :class:`.SearchResultEvent` event

        :param room: Room object or name to query
        :param query: The search query
        :return: An object containing the search request details and results
        """
        room_name = room.name if isinstance(room, Room) else room
        ticket = next(self._ticket_generator)

        await self._network.send_server_messages(
            RoomSearch.Request(room_name, ticket, query)
        )
        request = SearchRequest(
            ticket=ticket,
            query=query,
            search_type=SearchType.ROOM,
            room=room_name
        )
        await self._attach_request_timer_and_emit(request)

        return request

    async def search_user(self, username: str, query: str) -> SearchRequest:
        """Performs a search request on the specific user. The results generated
        by this query will stored in the returned object or can be listened to
        through the :class:`.SearchResultEvent` event

        :param user: User object or name to query
        :param query: The search query
        :return: An object containing the search request details and results
        """
        ticket = next(self._ticket_generator)

        await self._network.send_server_messages(
            UserSearch.Request(username, ticket, query)
        )
        request = SearchRequest(
            ticket=ticket,
            query=query,
            search_type=SearchType.USER,
            username=username
        )
        await self._attach_request_timer_and_emit(request)

        return request

    async def _query_shares_and_reply(self, ticket: int, username: str, query: str):
        """Performs a query on the shares manager and reports the results to the
        user
        """
        if not self._session:
            logger.warning("not returning search results : no valid session was set")
            return

        if self._settings.users.is_blocked(username, BlockingFlag.SEARCHES):
            return

        visible, locked = self._shares_manager.query(
            query,
            username=username,
            excluded_search_phrases=self.excluded_search_phrases
        )

        result_count = len(visible) + len(locked)
        self.received_searches.append(
            ReceivedSearch(
                username=username,
                query=query,
                result_count=result_count
            )
        )
        await self._event_bus.emit(
            SearchRequestReceivedEvent(
                username=username,
                query=query,
                result_count=result_count
            )
        )

        if len(visible) + len(locked) == 0:
            return

        logger.info(
            "found %d/%d results for query %r (username=%r)",
            len(visible), len(locked), query, username
        )

        task = asyncio.create_task(
            self._network.send_peer_messages(
                username,
                PeerSearchReply.Request(
                    username=self._session.user.name,
                    ticket=ticket,
                    results=convert_items_to_file_data(visible, use_full_path=True),
                    has_slots_free=self._upload_info_provider.has_slots_free(),
                    avg_speed=int(self._upload_info_provider.get_average_upload_speed()),
                    queue_size=self._upload_info_provider.get_queue_size(),
                    locked_results=convert_items_to_file_data(locked, use_full_path=True)
                )
            ),
            name=f'search-reply-{task_counter()}'
        )
        task.add_done_callback(
            partial(self._search_reply_task_callback, ticket, username, query))
        self._search_reply_tasks.append(task)

    def _search_reply_task_callback(self, ticket: int, username: str, query: str, task: asyncio.Task):
        """Callback for a search reply task. This callback simply logs the
        results and removes the task from the list
        """
        try:
            task.result()

        except asyncio.CancelledError:
            logger.debug(
                "cancelled delivery of search results (ticket=%d, username=%s, query=%s)",
                ticket, username, query
            )
        except Exception as exc:
            logger.warning(
                "failed to deliver search results : %r (ticket=%d, username=%s, query=%s)",
                exc, ticket, username, query
            )
        else:
            logger.info(
                "delivered search results (ticket=%d, username=%s, query=%s)",
                ticket, username, query
            )
        finally:
            self._search_reply_tasks.remove(task)

    async def _wishlist_job(self):
        """Job handling wishlist queries, this method is intended to be run as
        a task. This method will run at the given ``interval`` (returned by the
        server after logon).
        """
        items = self._settings.searches.wishlist

        timeout = self._get_wishlist_request_timeout()

        enabled_items = list(filter(lambda item: item.enabled, items))
        logger.info("starting wishlist search of %d items", len(enabled_items))
        # Recreate
        for item in enabled_items:
            ticket = next(self._ticket_generator)

            await self._network.send_server_messages(
                WishlistSearch.Request(ticket, item.query)
            )

            request = SearchRequest(
                ticket,
                item.query,
                search_type=SearchType.WISHLIST
            )
            request.timer = Timer(
                timeout=timeout,
                callback=partial(self._timeout_search_request, request)
            ) if timeout else None
            self.requests[ticket] = request

            if request.timer:
                request.timer.start()

            await self._event_bus.emit(SearchRequestSentEvent(request))

    def _get_wishlist_request_timeout(self) -> int:
        timeout = self._settings.searches.send.wishlist_request_timeout
        if self._settings.searches.send.wishlist_request_timeout < 0:
            if self.wishlist_interval is None:
                timeout = DEFAULT_WISHLIST_INTERVAL
            else:
                timeout = self.wishlist_interval

        return timeout

    async def _attach_request_timer_and_emit(self, request: SearchRequest):
        self.requests[request.ticket] = request

        if self._settings.searches.send.request_timeout > 0:
            request.timer = Timer(
                timeout=self._settings.searches.send.request_timeout,
                callback=partial(self._timeout_search_request, request)
            )
            request.timer.start()

        await self._event_bus.emit(SearchRequestSentEvent(request))

    async def _timeout_search_request(self, request: SearchRequest):
        del self.requests[request.ticket]
        await self._event_bus.emit(SearchRequestRemovedEvent(request))

    async def _on_message_received(self, event: MessageReceivedEvent):
        message = event.message
        if message.__class__ in self._MESSAGE_MAP:
            await self._MESSAGE_MAP[message.__class__](message, event.connection)

    @on_message(SearchInactivityTimeout.Response)
    async def _on_search_inactivity_timeout(self, message: SearchInactivityTimeout.Response, connection):
        self.search_inactivity_timeout = message.timeout

    @on_message(DistributedSearchRequest.Request)
    async def _on_distributed_search_request(
            self, message: DistributedSearchRequest.Request, connection: PeerConnection):

        await self._query_shares_and_reply(message.ticket, message.username, message.query)

    @on_message(DistributedServerSearchRequest.Request)
    async def _on_distributed_server_search_request(
            self, message: DistributedServerSearchRequest.Request, connection: PeerConnection):

        if message.distributed_code != DistributedSearchRequest.Request.MESSAGE_ID:
            logger.warning("no handling for server search request with code %d", message.distributed_code)
            return

        await self._query_shares_and_reply(message.ticket, message.username, message.query)

    @on_message(ServerSearchRequest.Response)
    async def _on_server_search_request(self, message: ServerSearchRequest.Response, connection: ServerConnection):
        if self._session is None:
            return

        username = self._session.user.name
        if message.username == username:
            return

        await self._query_shares_and_reply(
            message.ticket, message.username, message.query)

    @on_message(FileSearch.Response)
    async def _on_file_search(self, message: FileSearch.Response, connection: ServerConnection):
        """Received when user performs a user or room search"""
        if self._session is None:
            return

        if message.username == self._session.user.name:
            return

        await self._query_shares_and_reply(
            message.ticket, message.username, message.query)

    @on_message(PeerSearchReply.Request)
    async def _on_peer_search_reply(self, message: PeerSearchReply.Request, connection: PeerConnection):
        search_result = SearchResult(
            ticket=message.ticket,
            username=message.username,
            has_free_slots=message.has_slots_free,
            avg_speed=message.avg_speed,
            queue_size=message.queue_size,
            shared_items=message.results,
            locked_results=message.locked_results or []
        )
        try:
            query = self.requests[message.ticket]

        except KeyError:
            logger.warning("search reply ticket does not match any search request : %d", message.ticket)

        else:
            if self._settings.searches.send.store_results:
                query.results.append(search_result)

            await self._event_bus.emit(SearchResultEvent(query, search_result))

        await connection.disconnect(reason=CloseReason.REQUESTED)

    @on_message(WishlistInterval.Response)
    async def _on_wish_list_interval(self, message: WishlistInterval.Response, connection: ServerConnection):
        self.wishlist_interval = message.interval

        if task := self._wishlist_task.cancel():
            await task

        self._wishlist_task.interval = self.wishlist_interval
        self._wishlist_task.start()

    @on_message(ExcludedSearchPhrases.Response)
    async def _on_excluded_search_phrases(self, message: ExcludedSearchPhrases.Response, connection: ServerConnection):
        self.excluded_search_phrases = message.phrases

    async def _on_state_changed(self, event: ConnectionStateChangedEvent):
        if not isinstance(event.connection, ServerConnection):
            return

        if event.state == ConnectionState.CLOSING:
            self._wishlist_task.cancel()

    async def _on_session_initialized(self, event: SessionInitializedEvent):
        self._session = event.session

    async def _on_session_destroyed(self, event: SessionDestroyedEvent):
        self._session = None

    async def stop(self) -> list[asyncio.Task]:
        """Cancels all pending tasks

        :return: a list of tasks that have been cancelled so that they can be
            awaited
        """
        cancelled_tasks = []

        for task in self._search_reply_tasks:
            task.cancel()
            cancelled_tasks.append(task)

        if wishlist_task := self._wishlist_task.cancel():
            cancelled_tasks.append(wishlist_task)

        return cancelled_tasks



import re


DEFAULT_LISTENING_HOST: str = '0.0.0.0'
DEFAULT_PARENT_MIN_SPEED: int = 1
DEFAULT_PARENT_SPEED_RATIO: int = 50
DEFAULT_WISHLIST_INTERVAL: int = 600
DEFAULT_READ_TIMEOUT: float = 60
PEER_CONNECT_TIMEOUT: float = 10
"""Direct connection timeout"""
PEER_INDIRECT_CONNECT_TIMEOUT: float = 60
"""Indirect connection timeout"""
PEER_INIT_TIMEOUT: float = 5
"""Timeout waiting for Peer Initialization message"""
PEER_READ_TIMEOUT: float = 60
"""Timeout waiting for message on a peer or distributed connection"""
TRANSFER_TIMEOUT: float = 180
"""Timeout waiting for new data on a file connection"""
TRANSFER_REPLY_TIMEOUT = 30
SERVER_CONNECT_TIMEOUT: float = 30
SERVER_LOGIN_TIMEOUT: float = 30
SERVER_PING_INTERVAL: float = 5 * 60
SERVER_READ_TIMEOUT: float = SERVER_PING_INTERVAL * 2
DISCONNECT_TIMEOUT: float = 5
PATH_SEPERATOR_PATTERN = re.compile(r"[\\/]+")
"""Pattern for splitting/normalizing remote paths"""
UPNP_DEFAULT_CHECK_INTERVAL: int = 600
UPNP_DEFAULT_LEASE_DURATION: int = 6 * 60 * 60
UPNP_DEFAULT_SEARCH_TIMEOUT: int = 10
UPNP_MAPPING_SERVICES: list[str] = ["WANIPC", "WANPPP"]
POTENTIAL_PARENTS_CACHE_SIZE: int = 20
"""Maximum amount of potential parents stored"""
DEFAULT_COMMAND_TIMEOUT: float = 10
MIN_TRANSFER_MGMT_INTERVAL: float = 0.05
MAX_TRANSFER_MGMT_INTERVAL: float = 0.25



from collections.abc import Collection
import logging
from typing import Union
from .exceptions import AioSlskException
from .protocol.primitives import MessageDataclass
from .protocol import messages


def resolve(message_class: str) -> type[MessageDataclass]:
    message_root, req_resp = message_class.split('.')

    if root_message := getattr(messages, message_root, None):
        if req_resp_message := getattr(root_message, req_resp, None):
            return req_resp_message
        else:
            raise AioSlskException(
                f"could not resolve message with name : {req_resp} : {message_class}")
    else:
        raise AioSlskException(
            f"could not resolve message with name : {message_root} : {message_class}")


class MessageFilter(logging.Filter):
    """Logging filter for protocol messages"""

    def __init__(self, message_types: Collection[Union[type[MessageDataclass], str]]):
        super().__init__()
        converted = []
        for message_type in message_types:
            if isinstance(message_type, str):
                mtype = resolve(message_type)
            else:
                mtype = message_type
            converted.append(mtype)

        self.message_types: Collection[type[MessageDataclass]] = converted

    def filter(self, record: logging.LogRecord) -> bool:
        if message_type := getattr(record, 'message_type', None):
            return message_type not in self.message_types

        return True


class DistributedSearchMessageFilter(MessageFilter):
    """Logging filter that filters out any distributed search request messages
    """

    TYPES = (
        messages.ServerSearchRequest.Response,
        messages.DistributedSearchRequest.Request,
        messages.DistributedServerSearchRequest.Request,
    )

    def __init__(self):  # pragma: no cover
        super().__init__(self.TYPES)


class ConnectionLoggerAdapter(logging.LoggerAdapter):

    def process(self, msg, kwargs):
        if 'extra' in kwargs:
            extra = kwargs['extra']
            infos = [
                f"{extra['hostname']}:{extra['port']}"
            ]

            if 'connection_type' in extra:
                infos.append(extra['connection_type'])

            if 'username' in extra:
                infos.append(extra['username'] or 'N/A')

            return f"[{'|'.join(infos)}] {msg}", kwargs

        return super().process(msg, kwargs)



from __future__ import annotations
from functools import partial
from ipaddress import IPv4Address
import logging
from async_upnp_client.aiohttp import AiohttpRequester
from async_upnp_client.client_factory import UpnpFactory
from async_upnp_client.exceptions import UpnpActionResponseError
from async_upnp_client.profiles.igd import IgdDevice, PortMappingEntry
from async_upnp_client.search import async_search
from async_upnp_client.ssdp import SSDP_IP_V4, SSDP_PORT

from ..constants import (
    UPNP_DEFAULT_SEARCH_TIMEOUT,
    UPNP_MAPPING_SERVICES,
    UPNP_DEFAULT_LEASE_DURATION,
)

logger = logging.getLogger(__name__)


class UPNP:

    def __init__(self):
        self._factory: UpnpFactory = UpnpFactory(AiohttpRequester())

    async def search_igd_devices(self, source_ip: str, timeout: int = UPNP_DEFAULT_SEARCH_TIMEOUT) -> list[IgdDevice]:
        devices: list[IgdDevice] = []
        logger.info("starting search for IGD devices")
        await async_search(
            partial(self._search_callback, devices),
            source=(source_ip, 0),
            target=(SSDP_IP_V4, SSDP_PORT),
            timeout=timeout
        )
        logger.info("found %d IGD devices", len(devices))
        return devices

    async def _search_callback(self, devices: list[IgdDevice], headers):
        if headers['ST'] not in IgdDevice.DEVICE_TYPES:
            return

        logger.info("found Internet Gateway Device : %r", headers)
        device = await self._factory.async_create_device(headers['LOCATION'])

        devices.append(IgdDevice(device, None))

    async def get_mapped_ports(self, device: IgdDevice) -> list[PortMappingEntry]:
        entry_count = await device.async_get_port_mapping_number_of_entries()

        if entry_count:
            logger.debug("found %d mapped ports on device %r", entry_count, device.name)
            return await self._get_mapped_ports_known(device, entry_count)

        else:
            return await self._get_mapped_ports_unknown(device)

    async def _get_mapped_ports_known(self, device: IgdDevice, count: int) -> list[PortMappingEntry]:
        entries = []
        for idx in range(count):
            logger.debug("getting port map with index %d on device %r", idx, device.name)
            try:
                entry = await device.async_get_generic_port_mapping_entry(idx)

            except Exception as exc:
                logger.debug("failed to get entry %d on device : %r", idx, device.name, exc_info=exc)

            else:
                logger.debug("got entry for index %d on device %r : %r", idx, device.name, entry)
                if entry:
                    entries.append(entry)

        return entries

    async def _get_mapped_ports_unknown(self, device: IgdDevice) -> list[PortMappingEntry]:
        """Gets all mapped port entries for the given device when the length of
        the total amount of ports is not known.
        """
        entries = []
        idx = 0
        while True:
            logger.debug("getting port map with index %d on device %r", idx, device.name)
            try:
                entry = await device.async_get_generic_port_mapping_entry(idx)

            except UpnpActionResponseError as exc:
                # SpecifiedArrayIndexInvalid
                if not (exc.status == 500 and exc.error_code == 713):
                    logger.debug("failed to get entry %d on device : %r", idx, device.name, exc_info=exc)

                break

            except Exception as exc:
                logger.debug("failed to get entry %d on device : %r", idx, device.name, exc_info=exc)
                break

            else:
                if entry is None:
                    break
                logger.debug("got entry for index %d on device %r : %r", idx, device.name, entry)
                entries.append(entry)
                idx += 1

        return entries

    async def map_port(
            self, device: IgdDevice, internal_ip: str, port: int,
            lease_duration: int = UPNP_DEFAULT_LEASE_DURATION):

        action = device._any_action(UPNP_MAPPING_SERVICES, "AddPortMapping")
        if not action:
            raise ValueError(f"device {device.name} has no 'AddPortMapping' service")

        await action.async_call(
            NewRemoteHost='',
            NewExternalPort=port,
            NewProtocol='TCP',
            NewInternalPort=port,
            NewInternalClient=IPv4Address(internal_ip).exploded,
            NewEnabled=True,
            NewPortMappingDescription='AioSlsk',
            NewLeaseDuration=lease_duration
        )

    async def unmap_port(self, device: IgdDevice, port: int):
        action = device._any_action(UPNP_MAPPING_SERVICES, "DeletePortMapping")
        if not action:
            raise ValueError(f"device {device.name} has no 'DeletePortMapping' service")

        await action.async_call(
            NewRemoteHost='',
            NewExternalPort=port,
            NewProtocol='TCP'
        )


async def _main(args):
    upnp = UPNP()
    if args.subcommand == 'list':
        devices = await upnp.search_igd_devices(
            args.internal_ip, timeout=args.search_timeout)
        if not devices:
            print("No devices found")

        for device in devices:
            print(f"device : {device.name}")
            mapped_ports = await upnp.get_mapped_ports(device)
            for mapped_port in mapped_ports:
                print(mapped_port)

    if args.subcommand == 'map':
        devices = await upnp.search_igd_devices(
            args.internal_ip, timeout=args.search_timeout)
        if not devices:
            print("No devices found")

        for device in devices:
            await upnp.map_port(device, args.internal_ip, args.port)

    elif args.subcommand == 'unmap':
        devices = await upnp.search_igd_devices(
            args.internal_ip, timeout=args.search_timeout)
        if not devices:
            print("No devices found")

        for device in devices:
            await upnp.unmap_port(device, args.port)


if __name__ == '__main__':
    import asyncio
    import argparse
    import logging

    logging.basicConfig(level=logging.DEBUG)

    parser = argparse.ArgumentParser(description="Utility for mapping and unmapping ports to UPnP")
    subparsers = parser.add_subparsers(dest='subcommand')

    list_parser = subparsers.add_parser('list', help="List port mappings")
    list_parser.add_argument('internal_ip')
    list_parser.add_argument(
        '-st', '--search-timeout',
        type=int, default=5,
        help="UPnP Device search timeout"
    )

    map_parser = subparsers.add_parser('map', help="Add port mapping")
    map_parser.add_argument('internal_ip')
    map_parser.add_argument('port', type=int)
    map_parser.add_argument(
        '-st', '--search-timeout',
        type=int, default=5,
        help="UPnP Device search timeout"
    )

    unmap_parser = subparsers.add_parser('unmap', help="Remove port mapping")
    unmap_parser.add_argument('internal_ip')
    unmap_parser.add_argument('port', type=int)
    unmap_parser.add_argument(
        '-st', '--search-timeout',
        type=int, default=5,
        help="UPnP Device search timeout"
    )

    args = parser.parse_args()
    asyncio.run(_main(args))



import asyncio

from ..base_manager import BaseManager
from ..events import (
    build_message_map,
    on_message,
    EventBus,
    GlobalRecommendationsEvent,
    ItemRecommendationsEvent,
    ItemSimilarUsersEvent,
    MessageReceivedEvent,
    RecommendationsEvent,
    SessionInitializedEvent,
    SimilarUsersEvent,
    UserInterestsEvent,
)
from ..protocol.messages import (
    AddHatedInterest,
    AddInterest,
    GetGlobalRecommendations,
    GetItemRecommendations,
    GetItemSimilarUsers,
    GetRecommendations,
    GetSimilarUsers,
    GetUserInterests,
)
from ..network.connection import ServerConnection
from ..network.network import Network
from ..settings import Settings
from ..user.manager import UserManager


class InterestManager(BaseManager):
    """Class handling interests and recommendations"""

    def __init__(
            self, settings: Settings,
            event_bus: EventBus,
            user_manager: UserManager, network: Network):
        self._settings: Settings = settings
        self._event_bus: EventBus = event_bus
        self._user_manager: UserManager = user_manager
        self._network: Network = network

        self._MESSAGE_MAP = build_message_map(self)

        self.register_listeners()

    def register_listeners(self):
        self._event_bus.register(
            MessageReceivedEvent, self._on_message_received)
        self._event_bus.register(
            SessionInitializedEvent, self._on_session_initialized)

    async def advertise_interests(self):
        """Advertises all interests and hated interests defined in the settings
        to the server
        """
        messages = []
        for interest in self._settings.interests.liked:
            messages.append(
                self._network.send_server_messages(
                    AddInterest.Request(interest)
                )
            )

        for hated_interest in self._settings.interests.hated:
            messages.append(
                self._network.send_server_messages(
                    AddHatedInterest.Request(hated_interest)
                )
            )

        await asyncio.gather(*messages, return_exceptions=True)

    # Recommendations / interests
    @on_message(GetRecommendations.Response)
    async def _on_get_recommendations(
            self, message: GetRecommendations.Response, connection: ServerConnection):

        await self._event_bus.emit(
            RecommendationsEvent(
                recommendations=message.recommendations,
                unrecommendations=message.unrecommendations,
                raw_message=message
            )
        )

    @on_message(GetGlobalRecommendations.Response)
    async def _on_get_global_recommendations(
            self, message: GetGlobalRecommendations.Response, connection: ServerConnection):

        await self._event_bus.emit(
            GlobalRecommendationsEvent(
                recommendations=message.recommendations,
                unrecommendations=message.unrecommendations,
                raw_message=message
            )
        )

    @on_message(GetItemRecommendations.Response)
    async def _on_get_item_recommendations(
            self, message: GetItemRecommendations.Response, connection: ServerConnection):

        await self._event_bus.emit(
            ItemRecommendationsEvent(
                item=message.item,
                recommendations=message.recommendations,
                raw_message=message
            )
        )

    @on_message(GetUserInterests.Response)
    async def _on_get_user_interests(self, message: GetUserInterests.Response, connection: ServerConnection):
        await self._event_bus.emit(
            UserInterestsEvent(
                user=self._user_manager.get_user_object(message.username),
                interests=message.interests,
                hated_interests=message.hated_interests,
                raw_message=message
            )
        )

    @on_message(GetSimilarUsers.Response)
    async def _on_get_similar_users(self, message: GetSimilarUsers.Response, connection: ServerConnection):
        await self._event_bus.emit(
            SimilarUsersEvent(
                users=[
                    (self._user_manager.get_user_object(user.username), user.score, )
                    for user in message.users
                ],
                raw_message=message
            )
        )

    @on_message(GetItemSimilarUsers.Response)
    async def _on_get_item_similar_users(
            self, message: GetItemSimilarUsers.Response, connection: ServerConnection):
        await self._event_bus.emit(
            ItemSimilarUsersEvent(
                item=message.item,
                users=[
                    self._user_manager.get_user_object(username)
                    for username in message.usernames
                ],
                raw_message=message
            )
        )

    # Listeners

    async def _on_message_received(self, event: MessageReceivedEvent):
        message = event.message
        if message.__class__ in self._MESSAGE_MAP:
            await self._MESSAGE_MAP[message.__class__](message, event.connection)

    async def _on_session_initialized(self, event: SessionInitializedEvent):
        await self.advertise_interests()



from collections import OrderedDict
import logging
import time

from ..base_manager import BaseManager
from ..events import (
    build_message_map,
    ConnectionStateChangedEvent,
    EventBus,
    MessageReceivedEvent,
    on_message,
    PublicMessageEvent,
    RoomJoinedEvent,
    RoomLeftEvent,
    RoomListEvent,
    RoomMembersEvent,
    RoomMembershipGrantedEvent,
    RoomMembershipRevokedEvent,
    RoomMessageEvent,
    RoomOperatorGrantedEvent,
    RoomOperatorRevokedEvent,
    RoomOperatorsEvent,
    RoomTickerAddedEvent,
    RoomTickerRemovedEvent,
    RoomTickersEvent,
    SessionInitializedEvent,
)
from ..protocol.messages import (
    RoomChatMessage,
    JoinRoom,
    LeaveRoom,
    RoomTickers,
    RoomTickerAdded,
    RoomTickerRemoved,
    UserJoinedRoom,
    UserLeftRoom,
    PrivateRoomGrantMembership,
    PrivateRoomRevokeMembership,
    PrivateRoomMembers,
    PrivateRoomMembershipGranted,
    PrivateRoomOperators,
    PrivateRoomOperatorGranted,
    PrivateRoomOperatorRevoked,
    PrivateRoomGrantOperator,
    PrivateRoomRevokeOperator,
    PrivateRoomMembershipRevoked,
    PublicChatMessage,
    TogglePrivateRoomInvites,
    RoomList,
)
from .model import Room, RoomMessage
from ..network.connection import ConnectionState, ServerConnection
from ..network.network import Network
from ..settings import Settings
from ..user.manager import UserManager
from ..user.model import BlockingFlag, UserStatus


logger = logging.getLogger(__name__)


class RoomManager(BaseManager):
    """Class handling rooms"""

    def __init__(
            self, settings: Settings, event_bus: EventBus,
            user_manager: UserManager, network: Network):

        self._settings: Settings = settings
        self._event_bus: EventBus = event_bus
        self._user_manager: UserManager = user_manager
        self._network: Network = network

        self._rooms: dict[str, Room] = dict()

        self._MESSAGE_MAP = build_message_map(self)

        self.register_listeners()

    @property
    def rooms(self) -> dict[str, Room]:
        return self._rooms

    def get_joined_rooms(self) -> list[Room]:
        """Returns a list of all rooms currently joined"""
        return [room for room in self._rooms.values() if room.joined]

    def get_private_rooms(self) -> list[Room]:
        """Returns a list of all private rooms of which we are a member or
        owner. This includes rooms that are not joined
        """
        return [room for room in self._rooms.values() if room.private]

    def get_public_rooms(self) -> list[Room]:
        """Returns the list of public rooms"""
        return [room for room in self._rooms.values() if not room.private]

    def get_owned_rooms(self) -> list[Room]:
        """Returns a list of which we are the owner"""
        me = self._user_manager.get_self()
        return [
            room for room in self._rooms.values()
            if room.owner == me.name
        ]

    def get_operated_rooms(self) -> list[Room]:
        """Returns a list of rooms in which we have operator privileges"""
        me = self._user_manager.get_self()
        return [
            room for room in self._rooms.values()
            if me.name in room.operators
        ]

    def get_or_create_room(self, room_name: str, private: bool = False) -> Room:
        try:
            return self._rooms[room_name]
        except KeyError:
            room = Room(name=room_name, private=private)
            self._rooms[room_name] = room
            return room

    def reset_rooms(self):
        """Performs a reset on all users and rooms"""
        self._rooms = {}

    def register_listeners(self):
        self._event_bus.register(
            MessageReceivedEvent, self._on_message_received)
        self._event_bus.register(
            ConnectionStateChangedEvent, self._on_state_changed)
        self._event_bus.register(
            SessionInitializedEvent, self._on_session_initialized)

    async def auto_join_rooms(self):
        """Automatically joins rooms stored in the settings"""
        rooms = self._settings.rooms.favorites
        logger.info("automatically rejoining %d rooms", len(rooms))
        await self._network.send_server_messages(
            *[JoinRoom.Request(room) for room in rooms]
        )

    @on_message(RoomChatMessage.Response)
    async def _on_chat_room_message(self, message: RoomChatMessage.Response, connection: ServerConnection):

        if self._settings.users.is_blocked(message.username, BlockingFlag.ROOM_MESSAGES):
            return

        user = self._user_manager.get_user_object(message.username)
        room = self.get_or_create_room(message.room)
        room_message = RoomMessage(
            timestamp=int(time.time()),
            room=room,
            user=user,
            message=message.message
        )

        await self._event_bus.emit(
            RoomMessageEvent(
                room_message,
                raw_message=message
            )
        )

    @on_message(PublicChatMessage.Response)
    async def _on_public_chat_message(self, message: PublicChatMessage.Response, connection: ServerConnection):

        if self._settings.users.is_blocked(message.username, BlockingFlag.ROOM_MESSAGES):
            return

        room = self.get_or_create_room(message.room)
        user = self._user_manager.get_user_object(message.username)

        await self._event_bus.emit(
            PublicMessageEvent(
                timestamp=int(time.time()),
                room=room,
                user=user,
                message=message.message,
                raw_message=message
            )
        )

    @on_message(UserJoinedRoom.Response)
    async def _on_user_joined_room(self, message: UserJoinedRoom.Response, connection: ServerConnection):
        user = self._user_manager.get_user_object(message.username)
        user.status = UserStatus(message.status)
        user.update_from_user_stats(message.user_stats)
        user.slots_free = message.slots_free
        user.country = message.country_code

        room = self.get_or_create_room(message.room)
        room.add_user(user)

        await self._event_bus.emit(
            RoomJoinedEvent(
                room=room,
                user=user,
                raw_message=message
            )
        )

    @on_message(UserLeftRoom.Response)
    async def _on_user_left_room(self, message: UserLeftRoom.Response, connection: ServerConnection):
        user = self._user_manager.get_user_object(message.username)
        room = self.get_or_create_room(message.room)
        room.remove_user(user)

        await self._event_bus.emit(
            RoomLeftEvent(
                room=room,
                user=user,
                raw_message=message
            )
        )

    @on_message(JoinRoom.Response)
    async def _on_join_room(self, message: JoinRoom.Response, connection: ServerConnection):
        room = self.get_or_create_room(message.room)
        room.joined = True
        # There isn't a boolean to indicate that a room is private, but a private
        # room should always have an owner
        room.private = bool(message.owner)

        for idx, name in enumerate(message.users):
            user = self._user_manager.get_user_object(name)
            user.status = UserStatus(message.users_status[idx])
            user.update_from_user_stats(message.users_stats[idx])
            user.country = message.users_countries[idx]
            user.slots_free = message.users_slots_free[idx]

            room.add_user(user)

        # For private rooms
        room.owner = message.owner
        room.operators = set(message.operators or [])

        await self._event_bus.emit(
            RoomJoinedEvent(
                room=room,
                raw_message=message
            )
        )

    @on_message(LeaveRoom.Response)
    async def _on_leave_room(self, message: LeaveRoom.Response, connection: ServerConnection):
        """Received when we leave a room"""
        room = self.get_or_create_room(message.room)
        room.joined = False
        room.users = []
        await self._event_bus.emit(
            RoomLeftEvent(
                room=room,
                raw_message=message
            )
        )

    @on_message(RoomTickers.Response)
    async def _on_chat_room_tickers(self, message: RoomTickers.Response, connection: ServerConnection):
        room = self.get_or_create_room(message.room)
        tickers = OrderedDict()
        for ticker in message.tickers:
            self._user_manager.get_user_object(ticker.username)
            tickers[ticker.username] = ticker.ticker

        room.tickers = tickers

        await self._event_bus.emit(
            RoomTickersEvent(
                room=room,
                tickers=tickers,
                raw_message=message
            )
        )

    @on_message(RoomTickerAdded.Response)
    async def _on_chat_room_ticker_added(self, message: RoomTickerAdded.Response, connection: ServerConnection):
        room = self.get_or_create_room(message.room)
        user = self._user_manager.get_user_object(message.username)

        room.tickers[user.name] = message.ticker

        await self._event_bus.emit(
            RoomTickerAddedEvent(
                room,
                user,
                message.ticker,
                raw_message=message
            )
        )

    @on_message(RoomTickerRemoved.Response)
    async def _on_chat_room_ticker_removed(self, message: RoomTickerRemoved.Response, connection: ServerConnection):
        room = self.get_or_create_room(message.room)
        user = self._user_manager.get_user_object(message.username)

        try:
            del room.tickers[user.name]
        except KeyError:
            logger.warning(
                "attempted to remove room ticker for user %s in room %s but it wasn't present",
                user.name, room.name
            )

        await self._event_bus.emit(
            RoomTickerRemovedEvent(
                room,
                user,
                raw_message=message
            )
        )

    @on_message(TogglePrivateRoomInvites.Response)
    async def _on_private_room_toggle(
            self, message: TogglePrivateRoomInvites.Response, connection: ServerConnection):

        logger.debug("private rooms enabled : %s", message.enabled)

    @on_message(PrivateRoomGrantMembership.Response)
    async def _on_private_room_add_user(
            self, message: PrivateRoomGrantMembership.Response, connection: ServerConnection):

        room = self.get_or_create_room(message.room, private=True)
        user = self._user_manager.get_user_object(message.username)
        room.members.add(message.username)

        await self._event_bus.emit(
            RoomMembershipGrantedEvent(
                room=room,
                member=user,
                raw_message=message
            )
        )

    @on_message(PrivateRoomMembershipGranted.Response)
    async def _on_private_room_added(
            self, message: PrivateRoomMembershipGranted.Response, connection: ServerConnection):

        room = self.get_or_create_room(message.room, private=True)
        user = self._user_manager.get_self()
        room.members.add(user.name)

        await self._event_bus.emit(
            RoomMembershipGrantedEvent(
                room=room,
                raw_message=message
            )
        )

    @on_message(PrivateRoomRevokeMembership.Response)
    async def _on_private_room_remove_user(
            self, message: PrivateRoomRevokeMembership.Response, connection: ServerConnection):

        room = self.get_or_create_room(message.room, private=True)
        user = self._user_manager.get_user_object(message.username)

        room.members.discard(message.username)
        room.operators.discard(message.username)

        await self._event_bus.emit(
            RoomMembershipRevokedEvent(
                room=room,
                member=user,
                raw_message=message
            )
        )

    @on_message(PrivateRoomMembershipRevoked.Response)
    async def _on_private_room_removed(
            self, message: PrivateRoomMembershipRevoked.Response, connection: ServerConnection):

        room = self.get_or_create_room(message.room, private=True)
        user = self._user_manager.get_self()
        room.members.discard(user.name)
        room.operators.discard(user.name)

        await self._event_bus.emit(
            RoomMembershipRevokedEvent(
                room=room,
                raw_message=message
            )
        )

    @on_message(PrivateRoomMembers.Response)
    async def _on_private_room_users(self, message: PrivateRoomMembers.Response, connection: ServerConnection):
        room = self.get_or_create_room(message.room, private=True)
        room.members = set(message.usernames)

        await self._event_bus.emit(
            RoomMembersEvent(
                room=room,
                members=list(map(self._user_manager.get_user_object, message.usernames)),
                raw_message=message
            )
        )

    @on_message(PrivateRoomOperators.Response)
    async def _on_private_room_operators(
            self, message: PrivateRoomOperators.Response, connection: ServerConnection):

        room = self.get_or_create_room(message.room, private=True)
        room.operators = set(message.usernames)

        await self._event_bus.emit(
            RoomOperatorsEvent(
                room=room,
                operators=list(map(self._user_manager.get_user_object, room.operators)),
                raw_message=message
            )
        )

    @on_message(PrivateRoomOperatorGranted.Response)
    async def _on_operator_granted(
            self, message: PrivateRoomOperatorGranted.Response, connection: ServerConnection):

        room = self.get_or_create_room(message.room, private=True)
        room.operators.discard(self._user_manager.get_self().name)

        await self._event_bus.emit(
            RoomOperatorGrantedEvent(
                room=room,
                raw_message=message
            )
        )

    @on_message(PrivateRoomOperatorRevoked.Response)
    async def _on_operator_revoked(
            self, message: PrivateRoomOperatorRevoked.Response, connection: ServerConnection):

        room = self.get_or_create_room(message.room, private=True)
        room.operators.discard(self._user_manager.get_self().name)

        await self._event_bus.emit(
            RoomOperatorRevokedEvent(
                room=room,
                raw_message=message
            )
        )

    @on_message(PrivateRoomGrantOperator.Response)
    async def _on_user_operator_granted(
            self, message: PrivateRoomGrantOperator.Response, connection: ServerConnection):

        user = self._user_manager.get_user_object(message.username)
        room = self.get_or_create_room(message.room, private=True)
        room.operators.add(message.username)

        await self._event_bus.emit(
            RoomOperatorGrantedEvent(
                room=room,
                member=user,
                raw_message=message
            )
        )

    @on_message(PrivateRoomRevokeOperator.Response)
    async def _on_user_operator_revoked(
            self, message: PrivateRoomRevokeOperator.Response, connection: ServerConnection):

        user = self._user_manager.get_user_object(message.username)
        room = self.get_or_create_room(message.room, private=True)
        room.operators.discard(message.username)

        await self._event_bus.emit(
            RoomOperatorRevokedEvent(
                room=room,
                member=user,
                raw_message=message
            )
        )

    @on_message(RoomList.Response)
    async def _on_room_list(self, message: RoomList.Response, connection: ServerConnection):
        me = self._user_manager.get_self()
        for idx, room_name in enumerate(message.rooms):
            room = self.get_or_create_room(room_name, private=False)
            room.user_count = message.rooms_user_count[idx]

        for idx, room_name in enumerate(message.rooms_private_owned):
            room = self.get_or_create_room(room_name, private=True)
            room.owner = me.name
            room.user_count = message.rooms_private_owned_user_count[idx]

        for idx, room_name in enumerate(message.rooms_private):
            room = self.get_or_create_room(room_name, private=True)
            room.members.add(me.name)
            room.user_count = message.rooms_private_user_count[idx]

        for room_name in message.rooms_private_operated:
            room = self.get_or_create_room(room_name, private=True)
            room.operators.add(me.name)

        # Remove all rooms no longer tracked
        all_rooms = set(message.rooms) | set(message.rooms_private) | set(message.rooms_private_owned)
        unknown_rooms = set(self._rooms.keys()) - all_rooms
        for unknown_room in unknown_rooms:
            del self._rooms[unknown_room]

        # For the remaining rooms update owner, operators, members, private
        for room_name, room in self._rooms.items():
            if room_name not in message.rooms_private_owned:
                if room.owner == me.name:
                    room.owner = None

            if room_name not in message.rooms_private_operated:
                room.operators.discard(me.name)

            if room_name not in message.rooms_private:
                room.members.discard(me.name)

            room.private = room_name not in message.rooms

        await self._event_bus.emit(
            RoomListEvent(
                rooms=list(self._rooms.values()),
                raw_message=message
            )
        )

    # Listeners

    async def _on_message_received(self, event: MessageReceivedEvent):
        message = event.message
        if message.__class__ in self._MESSAGE_MAP:
            await self._MESSAGE_MAP[message.__class__](message, event.connection)

    async def _on_state_changed(self, event: ConnectionStateChangedEvent):
        if not isinstance(event.connection, ServerConnection):
            return

        if event.state == ConnectionState.CLOSED:
            self.reset_rooms()

    async def _on_session_initialized(self, event: SessionInitializedEvent):
        await self._network.send_server_messages(
            TogglePrivateRoomInvites.Request(self._settings.rooms.private_room_invites)
        )
        if not self._settings.rooms.auto_join:
            await self.auto_join_rooms()



import hashlib
import logging
import os
import shelve
from typing import Protocol
from .model import Transfer


logger = logging.getLogger(__name__)


class TransferCache(Protocol):
    """Abstract base class for storing shares"""

    def read(self) -> list['Transfer']:
        ...

    def write(self, transfers: list['Transfer']):
        ...


class TransferNullCache:
    """Transfer cache object that does not perform any caching"""

    def read(self) -> list['Transfer']:  # pragma: no cover
        return []

    def write(self, transfers: list['Transfer']):  # pragma: no cover
        pass


class TransferShelveCache:
    """Transfer cache that uses the Python built-in :py:mod:`shelve` module to
    store transfer objects
    """

    DEFAULT_FILENAME = 'transfers'

    def __init__(self, data_directory: str):
        self.data_directory = data_directory

    def read(self) -> list['Transfer']:
        db_path = os.path.join(self.data_directory, self.DEFAULT_FILENAME)

        transfers = []
        with shelve.open(db_path, flag='c') as database:
            for _, transfer in database.items():
                transfers.append(transfer)

        logger.info("read %d transfers from : %s", len(transfers), db_path)

        return transfers

    def write(self, transfers: list['Transfer']):
        db_path = os.path.join(self.data_directory, self.DEFAULT_FILENAME)

        logger.info("writing %d transfers to : %s", len(transfers), db_path)

        with shelve.open(db_path, flag='c') as database:
            # Update/add transfers
            for transfer in transfers:
                key = hashlib.sha256(
                    (
                        transfer.username +
                        transfer.remote_path +
                        str(transfer.direction.value)
                    ).encode('utf-8')
                ).hexdigest()
                database[key] = transfer

            # Remove non existing transfers
            keys_to_delete = []
            for key, db_transfer in database.items():
                if not any(transfer == db_transfer for transfer in transfers):
                    keys_to_delete.append(key)
            for key_to_delete in keys_to_delete:
                database.pop(key_to_delete)

        logger.info("successfully wrote %d transfers to : %s", len(transfers), db_path)
